from typing import Union
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Tuple
from typing import Any
from typing import Optional
from dataclasses import dataclass

import numpy as np


class Hypothesis(NamedTuple):
    """Hypothesis data type."""

    yseq: np.ndarray
    score: Union[float, np.ndarray] = 0
    scores: Dict[str, Union[float, np.ndarray]] = dict()
    states: Dict[str, Any] = dict()

    def asdict(self) -> dict:
        """Convert data to JSON-friendly dict."""
        return self._replace(
            yseq=self.yseq.tolist(),
            score=float(self.score),
            scores={k: float(v) for k, v in self.scores.items()},
        )._asdict()


class BatchHypothesis(NamedTuple):
    """Batchfied/Vectorized hypothesis data type."""

    yseq: np.ndarray = np.ndarray([])  # (batch, maxlen)
    score: np.ndarray = np.ndarray([])  # (batch,)
    length: np.ndarray = np.ndarray([])  # (batch,)
    scores: Dict[str, np.ndarray] = dict()  # values: (batch,)
    states: Dict[str, Dict] = dict()

    def __len__(self) -> int:
        """Return a batch size."""
        return len(self.length)


@dataclass
class TransducerHypothesis:
    """Default hypothesis definition for Transducer search algorithms."""

    score: float
    yseq: List[int]
    dec_state: Union[
        Tuple[np.ndarray, Optional[np.ndarray]],
        List[Optional[np.ndarray]],
        np.ndarray,
    ]
    lm_state: Union[Dict[str, Any], List[Any]] = None


@dataclass
class ExtendedHypothesis(TransducerHypothesis):
    """Extended hypothesis definition for NSC beam search and mAES."""

    dec_out: List[np.ndarray] = None
    lm_scores: np.ndarray = None



"""Ngram lm implement."""

from abc import ABC

import kenlm
import numpy as np

from .interface import BatchScorerInterface
from .interface import PartialScorerInterface


class Ngrambase(ABC):
    """Ngram base implemented through ScorerInterface."""

    def __init__(self, ngram_model, token_list):
        """Initialize Ngrambase.
        Args:
            ngram_model: ngram model path
            token_list: token list from dict or model.json
        """
        self.chardict = [x if x != "<eos>" else "</s>" for x in token_list]
        self.charlen = len(self.chardict)
        self.lm = kenlm.LanguageModel(ngram_model)
        self.tmpkenlmstate = kenlm.State()

    def init_state(self, x):
        """Initialize tmp state."""
        state = kenlm.State()
        self.lm.NullContextWrite(state)
        return state

    def score_partial_(self, y, next_token, state, x):
        """Score interface for both full and partial scorer.
        Args:
            y: previous char
            next_token: next token need to be score
            state: previous state
            x: encoded feature
        Returns:
            tuple[torch.Tensor, List[Any]]: Tuple of
                batchfied scores for next token with shape of `(n_batch, n_vocab)`
                and next state list for ys.
        """
        out_state = kenlm.State()
        ys = self.chardict[y[-1]] if y.shape[0] > 1 else "<s>"
        self.lm.BaseScore(state, ys, out_state)
        scores = np.zeros(next_token, dtype=x.dtype)
        for i, j in enumerate(next_token):
            scores[i] = self.lm.BaseScore(
                out_state, self.chardict[j], self.tmpkenlmstate
            )
        return scores, out_state


class NgramFullScorer(Ngrambase, BatchScorerInterface):
    """Fullscorer for ngram."""

    def score(self, y, state, x):
        """Score interface for both full and partial scorer.
        Args:
            y: previous char
            state: previous state
            x: encoded feature
        Returns:
            tuple[torch.Tensor, List[Any]]: Tuple of
                batchfied scores for next token with shape of `(n_batch, n_vocab)`
                and next state list for ys.
        """
        return self.score_partial_(y, np.arange(self.charlen), state, x)


class NgramPartScorer(Ngrambase, PartialScorerInterface):
    """Partialscorer for ngram."""

    def score_partial(self, y, next_token, state, x):
        """Score interface for both full and partial scorer.
        Args:
            y: previous char
            next_token: next token need to be score
            state: previous state
            x: encoded feature
        Returns:
            tuple[torch.Tensor, List[Any]]: Tuple of
                batchfied scores for next token with shape of `(n_batch, n_vocab)`
                and next state list for ys.
        """
        return self.score_partial_(y, next_token, state, x)

    def select_state(self, state, i):
        """Empty select state for scorer interface."""
        return state



from typing import (
    Optional,
    Tuple
)
import six
import os

import math
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from espnet.nets.pytorch_backend.rnn.attentions import (
    AttForward,
    AttForwardTA
)

from espnet_onnx.export.asr.models.language_models.embed import Embedding
from espnet_onnx.utils.torch_function import MakePadMask
from espnet_onnx.utils.abs_model import AbsExportModel
from espnet_onnx.export.layers.predecoder import PreDecoder
from espnet_onnx.export.layers.attention import (
    get_attention,
    OnnxAttForward,
)


class OnnxEncoderLayer(nn.Module):
    def __init__(self, model):
        super().__init__()
        self.model = model
    
    def forward(self, xs):
        """Inference.
        Args:
            x (Tensor): The sequeunce of character ids (T,)
                    or acoustic feature (T, idim * encoder_reduction_factor).
        Returns:
            Tensor: The sequences of encoder states(T, eunits).
        """
        xs = self.model.embed(xs).transpose(1, 2)
        if self.model.convs is not None:
            for i in six.moves.range(len(self.model.convs)):
                if self.model.use_residual:
                    xs = xs + self.model.convs[i](xs)
                else:
                    xs = self.model.convs[i](xs)
                    
        if self.model.blstm is None:
            return xs.transpose(1, 2)
        
        xs, _ = self.model.blstm(xs.transpose(1, 2))  # (B, Tmax, C)
        return xs[0]


class OnnxTacotron2Encoder(nn.Module, AbsExportModel):
    def __init__(
        self,
        model,
        **kwargs
    ):
        super().__init__()
        self.model_name = 'tts_model_encoder'
        
        # HPs
        self.odim = model.odim
        self.use_gst = model.use_gst
        self.spks = model.spks
        self.langs = model.langs
        self.spk_embed_dim = model.spk_embed_dim
        self.eos = model.eos
        
        # models
        self.enc = OnnxEncoderLayer(model.enc)
        if self.use_gst:
            self.gst = model.gst
        if self.spks is not None:
            self.sid_emb = model.sid_emb
        if self.langs is not None:
            self.lid_emb = model.lid_emb
        if self.spk_embed_dim is not None:
            self.spk_embed_integration_type = model.spk_embed_integration_type
            if self.spk_embed_integration_type == "add":
                self.projection = model.projection

    def _integrate_with_spk_embed(
        self, hs: torch.Tensor, spembs: torch.Tensor
    ) -> torch.Tensor:
        if self.spk_embed_integration_type == "add":
            # apply projection and then add to hidden states
            spembs = self.projection(F.normalize(spembs))
            hs = hs + spembs.unsqueeze(1)
        elif self.spk_embed_integration_type == "concat":
            # concat hidden states with spk embeds
            spembs = F.normalize(spembs).unsqueeze(1).expand(-1, hs.size(1), -1)
            hs = torch.cat([hs, spembs], dim=-1)
        else:
            raise NotImplementedError("support only add or concat.")

        return hs

    def forward(
        self,
        text: torch.Tensor,
        feats: Optional[torch.Tensor] = None,
        sids: Optional[torch.Tensor] = None,
        spembs: torch.Tensor = None,
        lids: Optional[torch.Tensor] = None,
    ):
        x = text
        y = feats
        spemb = spembs

        # add eos at the last of sequence
        x = F.pad(x, [0, 1], "constant", self.eos)

        # inference
        h = self.enc(x.unsqueeze(0))
        if self.use_gst:
            style_emb = self.gst(y.unsqueeze(0))
            h = h + style_emb
        if self.spks is not None:
            sid_emb = self.sid_emb(sids.view(-1))
            h = h + sid_emb
        if self.langs is not None:
            lid_emb = self.lid_emb(lids.view(-1))
            h = h + lid_emb
        if self.spk_embed_dim is not None:
            hs, spembs = h.unsqueeze(0), spemb.unsqueeze(0)
            h = self._integrate_with_spk_embed(hs, spembs)[0]
        
        return h
        
    def get_dummy_inputs(self):
        text = torch.LongTensor([0, 1])
        feats = torch.randn(10, self.odim) \
            if self.use_gst else None

        sids = torch.LongTensor([0]) \
            if self.spks is not None else None

        spembs = torch.randn(self.spk_embed_dim) \
            if self.spk_embed_dim is not None else None

        lids = torch.LongTensor([0]) \
            if self.langs is not None else None

        return (text, feats, sids, spembs, lids)

    def get_input_names(self):
        return ['text', 'feats', 'sids', 'spembs', 'lids']

    def get_output_names(self):
        return ['h']

    def get_dynamic_axes(self):
        ret = {
            'text': {0: 'text_length'},
            'feats': {0: 'feats_length'},
        }
        return ret

    def get_model_config(self, path):
        return {
            'model_type': 'Tacotron2Encoder',
            'model_path': str(path / f'{self.model_name}.onnx'),
            'eos': self.eos,
        }


class PostDecoder(nn.Module, AbsExportModel):
    def __init__(self, model, activation, odim):
        super().__init__()
        self.model = model
        self.odim = odim
        self.activation = activation
        self.model_name = f'postdecoder'
    
    def forward(self, x):
        if self.model is not None:
            x = x + self.model(x)
        x = x.transpose(2, 1).squeeze(0)
        if self.activation is not None:
            x = self.activation(x)
        return x

    def get_dummy_inputs(self):
        x = torch.randn(1, self.odim, 100)
        return (x,)

    def get_input_names(self):
        return ['x']

    def get_output_names(self):
        return ['out']

    def get_dynamic_axes(self):
        return {
            'x': {
                2: 'x_length'
            }
        }

    def get_model_config(self, path):
        return {
            "model_path": os.path.join(path, f'{self.model_name}.onnx'),
        }


class OnnxTacotron2Decoder(nn.Module, AbsExportModel):
    def __init__(
        self,
        model,
        threshold=0.5,
        minlenratio=0.0,
        maxlenratio=10.0,
        use_att_constraint=False,
        backward_window=1,
        forward_window=3,
        **kwargs
    ):
        super().__init__()
        self.model = model
        self.submodel = []
        self.model_name = 'tts_model_decoder'
        self.model.eval()
        
        # HPs
        self.odim = model.odim
        self.enc_out = model.enc.blstm.hidden_size * 2
        self.threshold = threshold
        self.minlenratio = minlenratio
        self.maxlenratio = maxlenratio
        self.backward_window = backward_window
        self.forward_window = forward_window
        self.use_concate = model.dec.use_concate
        self.cumulate_att_w = model.cumulate_att_w
        self.use_att_extra_inputs = model.dec.use_att_extra_inputs
        self.use_att_constraint = use_att_constraint
        self.onnx_export = model.dec.postnet is not None
        self.reduction_factor = model.dec.reduction_factor
        
        self.prenet = model.dec.prenet
        self.lstm = model.dec.lstm
        self.feat_out = model.dec.feat_out
        self.prob_out = model.dec.prob_out
        self.output_activation_fn = model.dec.output_activation_fn
        # get attention model
        self.att = get_attention(model.dec.att)
        if self.att is None:
            # attention model is AttForward or AttforwardTA.
            if isinstance(model.dec.att, AttForward) or isinstance(model.dec.att, AttForwardTA):
                raise ValueError("AttForwardTA or AttForward is not currently supported")
            
        self.submodel.append(
            PreDecoder(model.dec.att, 0)
        )
        self.submodel.append(
            PostDecoder(
                model.dec.postnet,
                model.dec.output_activation_fn,
                self.odim
            )
        )

    def forward(
        self,
        c_prev,
        z_prev,
        a_prev,
        pre_compute_enc_h,
        enc_h,
        mask,
        prev_out,
    ):
        # decoder calculation
        if self.use_att_extra_inputs:
            att_c, att_w = self.att(
                z_prev[0],
                a_prev,
                pre_compute_enc_h,
                enc_h,
                mask,
                prev_out,
                # last_attended_idx=last_attended_idx,
                # backward_window=self.backward_window,
                # forward_window=self.forward_window,
            )
        else:
            att_c, att_w = self.att(
                z_prev[0],
                a_prev,
                pre_compute_enc_h,
                enc_h,
                mask,
                # last_attended_idx=last_attended_idx,
                # backward_window=self.backward_window,
                # forward_window=self.forward_window,
            )

        prenet_out = prev_out
        if self.prenet is not None:
            for i in six.moves.range(len(self.prenet.prenet)):
                prenet_out = self.prenet.prenet[i](prenet_out)
            
        xs = torch.cat([att_c, prenet_out], dim=1)
        new_z_cache = []
        new_c_cache = []
        _z_list, _c_list = self.lstm[0](xs, (z_prev[0], c_prev[0]))
        new_z_cache.append(_z_list)
        new_c_cache.append(_c_list)
        
        for i in six.moves.range(1, len(self.lstm)):
            _z_list, _c_list = self.lstm[i](
                _z_list, (z_prev[i], c_prev[i])
            )
            new_z_cache.append(_z_list)
            new_c_cache.append(_c_list)
        
        zcs = (
            torch.cat([_z_list, att_c], dim=1)
            if self.use_concate
            else _z_list
        )
        out = self.feat_out(zcs).view(1, self.odim, -1)  # [(1, odim, r), ...]
        prob = torch.sigmoid(self.prob_out(zcs))[0]  # [(r), ...]
        if self.output_activation_fn is not None:
            prev_out = self.output_activation_fn(out[:, :, -1])  # (1, odim)
        else:
            prev_out = out[:, :, -1]  # (1, odim)

        return (
            out,
            prob,
            att_w,
            prev_out,
            new_c_cache,
            new_z_cache,
        )
        
    def get_a_prev(self, feat_length, att):
        ret = torch.randn(1, feat_length)
        # if att.att_type == 'location2d':
        #     ret = torch.randn(1, att.att_win, feat_length)
        if att.att_type in ('coverage', 'coverage_location'):
            ret = torch.randn(1, 1, feat_length)
        return ret

    def get_dummy_inputs(self):
        feat_length = 50
        make_pad_mask = MakePadMask(torch.LongTensor([1024]))
        z_prev = [torch.randn(1, self.model.dec.lstm[i].hidden_size)
                  for i in range(len(self.model.dec.lstm))]
        a_prev = self.get_a_prev(feat_length, self.att)
        c_prev = [torch.randn(1, self.model.dec.lstm[i].hidden_size)
                  for i in range(len(self.model.dec.lstm))]
        pre_compute_enc_h = self.get_precompute_enc_h(feat_length)
        enc_h = torch.randn(1, feat_length, self.enc_out)
        mask = torch.from_numpy(np.where(make_pad_mask(
            torch.LongTensor([feat_length])) == 1, -float('inf'), 0)).type(torch.float32)
        prev_out = torch.zeros(1, self.odim)
        return (
            c_prev, z_prev,
            a_prev, pre_compute_enc_h,
            enc_h, mask, prev_out
        )
    
    def get_precompute_enc_h(self,feat_length):
        return torch.randn(1, feat_length, self.model.dec.att.mlp_enc.out_features)

    def get_input_names(self):
        ret = ['c_prev_%d' % i for i in range(len(self.model.dec.lstm))]
        ret += ['z_prev_%d' % i for i in range(len(self.model.dec.lstm))]
        ret += ['a_prev', 'pceh', 'enc_h', 'mask', 'prev_in']
        return ret

    def get_output_names(self):
        ret = ['out', 'prob', 'prev_att_w', 'prev_out']
        ret += ['c_list_%d' % i for i in range(len(self.model.dec.lstm))]
        ret += ['z_list_%d' % i for i in range(len(self.model.dec.lstm))]
        return ret

    def get_dynamic_axes(self):
        # input
        ret = {}
        ret.update({
            'a_prev': {
                self.att.get_dynamic_axes(): 'a_prev_length',
            }
        })
        ret.update({
            'pceh': {
                1: 'pceh_length',
            }
        })
        ret.update({
            'enc_h': {
                1: 'enc_h_length',
            }
        })
        ret.update({
            'mask': {
                0: 'mask_length',
                1: 'mask_height'
            }
        })
        return ret

    def get_model_config(self, path):
        ret = {
            'model_type': 'Tacotron2Decoder',
            "model_path": os.path.join(path, f'{self.model_name}.onnx'),
            "dlayers": len(self.model.dec.lstm),
            "odim": self.odim,
            "dunits": self.model.dec.lstm[0].hidden_size,
            "threshold": self.threshold,
            "minlenratio": self.minlenratio,
            "maxlenratio": self.maxlenratio,
            "reduction_factor": self.reduction_factor,
            "cumulate_att_w": self.cumulate_att_w,
            "predecoder": {
                "model_path": os.path.join(path, f'{self.submodel[0].model_name}.onnx'),
                "att_type": self.att.att_type
            },
            "postdecoder": {
                "model_path": os.path.join(path, f'{self.submodel[1].model_name}.onnx'),
                "onnx_export": self.onnx_export
            }
        }
        if hasattr(self.model, 'att_win'):
            ret.update(att_win=self.model.att_win)
        return ret



import torch
import torch.nn as nn

import numpy as np


class MakePadMask(nn.Module):
    def __init__(self, max_seq_len=512):
        super().__init__()
        self.mask_pad = torch.Tensor(1 - np.tri(max_seq_len)).type(torch.bool)
    
    def forward(self, lengths, xs=None, length_dim=-1, maxlen=None):
        """Make mask tensor containing indices of padded part.
        This implementation creates the same mask tensor with original make_pad_mask,
        which can be converted into onnx format.
        Dimension length of xs should be 2 or 3.
        """
        if length_dim == 0:
            raise ValueError("length_dim cannot be 0: {}".format(length_dim))

        if xs is not None and len(xs.shape) == 3:
            if length_dim == 1:
                lengths = lengths.unsqueeze(1).expand(
                    *xs.transpose(1, 2).shape[:2])
            else:
                lengths = lengths.unsqueeze(1).expand(*xs.shape[:2])

        if maxlen is not None:
            m = maxlen
        elif xs is not None:
            m = xs.shape[-1]
        else:
            m = torch.max(lengths)

        mask = self.mask_pad[lengths - 1][..., :m].type(torch.float32)

        if length_dim == 1:
            return mask.transpose(1, 2)
        else:
            return mask



"""Length bonus module."""
from typing import Any
from typing import List
from typing import Tuple

import numpy as np

from .interface import BatchScorerInterface


class LengthBonus(BatchScorerInterface):
    """Length bonus in beam search."""

    def __init__(self, n_vocab: int):
        """Initialize class.
        Args:
            n_vocab (int): The number of tokens in vocabulary for beam search
        """
        self.n = n_vocab

    def score(self, y, state, x):
        """Score new token.
        Args:
            y (np.ndarray): 1D torch.int64 prefix tokens.
            state: Scorer state for prefix tokens
            x (np.ndarray): 2D encoder feature that generates ys.
        Returns:
            tuple[np.ndarray, Any]: Tuple of
                torch.float32 scores for next token (n_vocab)
                and None
        """
        return np.array([1.0], dtype=x.dtype).repeat(self.n), None

    def batch_score(
        self, ys: np.ndarray, states: List[Any], xs: np.ndarray
    ) -> Tuple[np.ndarray, List[Any]]:
        """Score new token batch.
        Args:
            ys (np.ndarray): torch.int64 prefix tokens (n_batch, ylen).
            states (List[Any]): Scorer states for prefix tokens.
            xs (np.ndarray):
                The encoder feature that generates ys (n_batch, xlen, n_feat).
        Returns:
            tuple[np.ndarray, List[Any]]: Tuple of
                batchfied scores for next token with shape of `(n_batch, n_vocab)`
                and next state list for ys.
        """
        return (
            np.array([1.0], dtype=xs.dtype)
            .repeat(ys.shape[0] * self.n)
            .reshape(ys.shape[0] * self.n),
            None,
        )



from typing import (
    Union,
    List,
    Tuple
)

import numpy as np
import onnxruntime
from scipy.special import log_softmax

from espnet_onnx.asr.scorer.interface import BatchScorerInterface
from espnet_onnx.utils.config import Config


class SequentialRNNLM(BatchScorerInterface):
    """Sequential RNNLM.
    See also:
        https://github.com/pytorch/examples/blob/4581968193699de14b56527296262dd76ab43557/word_language_model/model.py
    """

    def __init__(
        self,
        config: Config,
        providers: List[str],
        use_quantized: bool = False
    ):
        if use_quantized:
            self.lm_session = onnxruntime.InferenceSession(
                config.quantized_model_path,
                providers=providers
            )
        else:
            self.lm_session = onnxruntime.InferenceSession(
                config.model_path,
                providers=providers
            )
        self.enc_output_names = ['y'] \
            + [d.name for d in self.lm_session.get_outputs() if 'hidden' in d.name]
        self.enc_in_cache_names = [
            d.name for d in self.lm_session.get_inputs() if 'hidden' in d.name]

        self.rnn_type = config.rnn_type
        self.nhid = config.nhid
        self.nlayers = config.nlayers

    def zero_state(self):
        """Initialize LM state filled with zero values."""
        if self.rnn_type == 'LSTM':
            h = np.zeros((self.nlayers, self.nhid), dtype=np.float32)
            c = np.zeros((self.nlayers, self.nhid), dtype=np.float32)
            state = h, c
        else:
            state = np.zeros((self.nlayers, self.nhid), dtype=np.float32)
        return state

    def score(
        self,
        y: np.ndarray,
        state: Union[np.ndarray, Tuple[np.ndarray, np.ndarray]],
        x: np.ndarray,
    ) -> Tuple[np.ndarray, Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]]:
        """Score new token.
        Args:
            y: 1D torch.int64 prefix tokens.
            state: Scorer state for prefix tokens
            x: 2D encoder feature that generates ys.
        Returns:
            Tuple of
                torch.float32 scores for next token (n_vocab)
                and next state for ys
        """
        input_dic = {'x': y[-1].reshape(1, 1)}
        input_dic.update({
            k: v for k, v in zip(self.enc_in_cache_names, state)
        })
        decoded, *new_state = self.lm_session.run(
            self.enc_output_names,
            input_dic
        )
        logp = log_softmax(decoded, axis=-1).reshape(-1)
        return logp, new_state

    def batch_score(
        self, ys: np.ndarray, states: np.ndarray, xs: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Score new token batch.
        Args:
            ys (np.ndarray): torch.int64 prefix tokens (n_batch, ylen).
            states (List[Any]): Scorer states for prefix tokens.
            xs (np.ndarray):
                The encoder feature that generates ys (n_batch, xlen, n_feat).
        Returns:
            tuple[np.ndarray, List[Any]]: Tuple of
                batchfied scores for next token with shape of `(n_batch, n_vocab)`
                and next state list for ys.
        """
        if states[0] is None:
            c = np.zeros((self.nlayers, 1, self.nhid), dtype=np.float32)
            states = (c,)
            if self.rnn_type == 'LSTM':
                states = (c, c)

        elif self.rnn_type == 'LSTM':
            # states: Batch x 2 x (Nlayers, Dim) -> 2 x (Nlayers, Batch, Dim)
            h = np.concatenate([h[:, None] for h, c in states], axis=1)
            c = np.concatenate([c[:, None] for h, c in states], axis=1)
            states = h, c

        else:
            # states: Batch x (Nlayers, Dim) -> (Nlayers, Batch, Dim)
            states = np.concatenate([states[:, None] for s in states], axis=1)

        input_dic = {'x': ys[:, -1:].astype(np.int64)}
        input_dic.update({
            k: v for k, v in zip(self.enc_in_cache_names, states)
        })
        decoded, *new_states = self.lm_session.run(
            self.enc_output_names,
            input_dic
        )
        decoded = decoded.squeeze(1)
        logp = log_softmax(decoded, axis=-1)
        # state: Change to batch first
        if self.rnn_type == 'LSTM':
            # h, c: (Nlayers, Batch, Dim)
            h, c = new_states
            # states: Batch x 2 x (Nlayers, Dim)
            new_states = [(h[:, i], c[:, i]) for i in range(h.shape[1])]
        else:
            # states: (Nlayers, Batch, Dim) -> Batch x (Nlayers, Dim)
            new_states = [new_states[:, i] for i in range(new_states.shape[1])]

        return logp, new_states



import os

import numpy as np
import torch
import torch.nn as nn

from espnet.nets.pytorch_backend.transformer.subsampling import Conv2dSubsampling
from espnet.nets.pytorch_backend.transformer.subsampling import Conv2dSubsampling2
from espnet.nets.pytorch_backend.transformer.subsampling import Conv2dSubsampling6
from espnet.nets.pytorch_backend.transformer.subsampling import Conv2dSubsampling8
from espnet2.asr.frontend.default import DefaultFrontend
from espnet2.layers.global_mvn import GlobalMVN
from espnet2.layers.utterance_mvn import UtteranceMVN

from espnet_onnx.utils.torch_function import MakePadMask
from ..language_models.embed import Embedding
from espnet_onnx.utils.abs_model import AbsExportModel


class XformerEncoder(nn.Module, AbsExportModel):
    def __init__(
        self,
        model,
        max_seq_len=512,
        feats_dim=80, 
        **kwargs
    ):
        super().__init__()
        self.embed = Embedding(model.embed, max_seq_len)
        self.model = model
        self.make_pad_mask = MakePadMask(max_seq_len)
        self.feats_dim = feats_dim
        self.model_name = 'xformer_encoder'

    def forward(self, feats, feats_length):
        mask = 1 - self.make_pad_mask(feats_length).unsqueeze(1)
        if (
            isinstance(self.model.embed, Conv2dSubsampling)
            or isinstance(self.model.embed, Conv2dSubsampling2)
            or isinstance(self.model.embed, Conv2dSubsampling6)
            or isinstance(self.model.embed, Conv2dSubsampling8)
        ):
            xs_pad, mask = self.embed(feats, mask)
            xs_pad = xs_pad[0]
        else:
            xs_pad = self.embed(feats)

        xs_pad, masks = self.model.encoders(xs_pad, mask)
        if isinstance(xs_pad, tuple):
            xs_pad = xs_pad[0]
        if self.model.normalize_before:
            xs_pad = self.model.after_norm(xs_pad)

        olens = masks.squeeze(1).sum(1)
        return xs_pad, olens, None

    def get_output_size(self):
        return self.model.encoders[0].size

    def get_dummy_inputs(self):
        feats = torch.randn(1, 100, self.feats_dim)
        feats_lengths = torch.LongTensor([feats.size(1)])
        return (feats, feats_lengths)

    def get_input_names(self):
        return ['feats', 'feats_length']

    def get_output_names(self):
        return ['encoder_out', 'encoder_out_lens']

    def get_dynamic_axes(self):
        return {
            'feats': {
                1: 'feats_length'
            },
            'encoder_out': {
                1: 'enc_out_length'
            }
        }

    def get_model_config(self, asr_model=None, path=None):
        ret = {}
        ret.update(
            enc_type='XformerEncoder',
            model_path=os.path.join(path, f'{self.model_name}.onnx'),
            is_vggrnn=False,
            frontend=self.get_frontend_config(asr_model.frontend),
            do_normalize=asr_model.normalize is not None,
            do_preencoder=asr_model.preencoder is not None,
            do_postencoder=asr_model.postencoder is not None
        )
        if ret['do_normalize']:
            ret.update(normalize=self.get_norm_config(
                asr_model.normalize, path))
        # Currently preencoder, postencoder is not supported.
        # if ret['do_preencoder']:
        #     ret.update(preencoder=get_preenc_config(self.model.preencoder))
        # if ret['do_postencoder']:
        #     ret.update(postencoder=get_postenc_config(self.model.postencoder))
        return ret

    def get_frontend_config(self, frontend):
        # currently only default config is supported.
        assert isinstance(
            frontend, DefaultFrontend), 'Currently only DefaultFrontend is supported.'

        stft_config = dict(
            n_fft=frontend.stft.n_fft,
            win_length=frontend.stft.win_length,
            hop_length=frontend.stft.hop_length,
            window=frontend.stft.window,
            center=frontend.stft.center,
            onesided=frontend.stft.onesided,
            normalized=frontend.stft.normalized,
        )
        logmel_config = frontend.logmel.mel_options
        logmel_config.update(log_base=frontend.logmel.log_base)
        return {
            "stft": stft_config,
            "logmel": logmel_config
        }

    def get_norm_config(self, normalize, path):
        if isinstance(normalize, GlobalMVN):
            return {
                "type": "gmvn",
                "norm_means": normalize.norm_means,
                "norm_vars": normalize.norm_vars,
                "eps": normalize.eps,
                "stats_file": str(path.parent / 'feats_stats.npz')
            }
        elif isinstance(normalize, UtteranceMVN):
            return {
                "type": "utterance_mvn",
                "norm_means": normalize.norm_means,
                "norm_vars": normalize.norm_vars,
                "eps": normalize.eps,
            }



import six

from typing import (
    Any,
    List,
    Tuple
)
from typeguard import check_argument_types

import numpy as np
import onnxruntime
from scipy.special import (
    logsumexp,
    log_softmax
)

from espnet_onnx.utils.config import Config
from .interface import BatchPartialScorerInterface


class CTCPrefixScore:
    """Compute CTC label sequence scores
    which is based on Algorithm 2 in WATANABE et al.
    "HYBRID CTC/ATTENTION ARCHITECTURE FOR END-TO-END SPEECH RECOGNITION,"
    but extended to efficiently compute the probablities of multiple labels
    simultaneously
    """

    def __init__(self, x: np.ndarray, blank: int, eos: float, xp: np):
        assert check_argument_types()
        self.xp = xp
        self.logzero = -10000000000.0
        self.blank = blank
        self.eos = eos
        self.input_length = len(x)
        self.x = x

    def initial_state(self):
        """Obtain an initial CTC state
        :return: CTC state
        """
        # initial CTC state is made of a frame x 2 tensor that corresponds to
        # r_t^n(<sos>) and r_t^b(<sos>), where 0 and 1 of axis=1 represent
        # superscripts n and b (non-blank and blank), respectively.
        r = self.xp.full((self.input_length, 2),
                         self.logzero, dtype=np.float32)
        r[0, 1] = self.x[0, self.blank]
        for i in six.moves.range(1, self.input_length):
            r[i, 1] = r[i - 1, 1] + self.x[i, self.blank]
        return r

    def __call__(self, y, cs, r_prev):
        """Compute CTC prefix scores for next labels
        :param y     : prefix label sequence
        :param cs    : array of next labels
        :param r_prev: previous CTC state
        :return ctc_scores, ctc_states
        """
        # initialize CTC states
        output_length = len(y) - 1  # ignore sos
        # new CTC states are prepared as a frame x (n or b) x n_labels tensor
        # that corresponds to r_t^n(h) and r_t^b(h).
        r = self.xp.ndarray((self.input_length, 2, len(cs)), dtype=np.float32)
        xs = self.x[:, cs]
        if output_length == 0:
            r[0, 0] = xs[0]
            r[0, 1] = self.logzero
        else:
            r[output_length - 1] = self.logzero

        # prepare forward probabilities for the last label
        r_sum = self.xp.logaddexp(
            r_prev[:, 0], r_prev[:, 1]
        )  # log(r_t^n(g) + r_t^b(g))
        last = y[-1]
        if output_length > 0 and last in cs:
            log_phi = self.xp.ndarray(
                (self.input_length, len(cs)), dtype=np.float32)
            for i in six.moves.range(len(cs)):
                log_phi[:, i] = r_sum if cs[i] != last else r_prev[:, 1]
        else:
            log_phi = r_sum

        # compute forward probabilities log(r_t^n(h)), log(r_t^b(h)),
        # and log prefix probabilities log(psi)
        start = max(output_length, 1)
        log_psi = r[start - 1, 0]
        for t in six.moves.range(start, self.input_length):
            r[t, 0] = self.xp.logaddexp(r[t - 1, 0], log_phi[t - 1]) + xs[t]
            r[t, 1] = (
                self.xp.logaddexp(r[t - 1, 0], r[t - 1, 1]
                                  ) + self.x[t, self.blank]
            )
            log_psi = self.xp.logaddexp(log_psi, log_phi[t - 1] + xs[t])

        # get P(...eos|X) that ends with the prefix itself
        eos_pos = self.xp.where(cs == self.eos)[0]
        if len(eos_pos) > 0:
            log_psi[eos_pos] = r_sum[-1]  # log(r_T^n(g) + r_T^b(g))

        # exclude blank probs
        blank_pos = self.xp.where(cs == self.blank)[0]
        if len(blank_pos) > 0:
            log_psi[blank_pos] = self.logzero

        # return the log prefix probability and CTC states, where the label axis
        # of the CTC states is moved to the first axis to slice it easily
        return log_psi, self.xp.rollaxis(r, 2)


class CTCPrefixScorer(BatchPartialScorerInterface):
    """Decoder interface wrapper for CTCPrefixScore."""

    def __init__(self, ctc: Config, eos: int, providers: List[str], use_quantized: bool = False):
        """Initialize class.
        Args:
            ctc (np.ndarray): The CTC implementation.
                For example, :class:`espnet.nets.pytorch_backend.ctc.CTC`
            eos (int): The end-of-sequence id.
        """
        assert check_argument_types()
        if use_quantized:
            self.ctc = onnxruntime.InferenceSession(
                ctc.quantized_model_path,
                providers=providers
            )
        else:
            self.ctc = onnxruntime.InferenceSession(
                ctc.model_path,
                providers=providers
            )
        self.eos = eos
        self.impl = None

    def init_state(self, x: np.ndarray):
        """Get an initial state for decoding.
        Args:
            x (np.ndarray): The encoded feature tensor
        Returns: initial state
        """
        x = self.ctc.run(["ctc_out"], {"x": x[None, :]})[0]
        logp = np.squeeze(x, axis=0)
        # TODO(karita): use CTCPrefixScoreTH
        self.impl = CTCPrefixScore(logp, 0, self.eos, np)
        return 0, self.impl.initial_state()

    def select_state(self, state, i, new_id=None):
        """Select state with relative ids in the main beam search.
        Args:
            state: Decoder state for prefix tokens
            i (int): Index to select a state in the main beam search
            new_id (int): New label id to select a state if necessary
        Returns:
            state: pruned state
        """
        if type(state) == tuple:
            if len(state) == 2:  # for CTCPrefixScore
                sc, st = state
                return sc[i], st[i]
            else:  # for CTCPrefixScoreTH (need new_id > 0)
                r, log_psi, f_min, f_max, scoring_idmap = state
                s = log_psi[i, new_id].repeat(log_psi.shape[1])
                if scoring_idmap is not None:
                    return r[:, :, i, scoring_idmap[i, new_id]], s, f_min, f_max
                else:
                    return r[:, :, i, new_id], s, f_min, f_max
        return None if state is None else state[i]

    def score_partial(self, y, ids, state, x):
        """Score new token.
        Args:
            y (np.ndarray): 1D prefix token
            next_tokens (np.ndarray): torch.int64 next token to score
            state: decoder state for prefix tokens
            x (np.ndarray): 2D encoder feature that generates ys
        Returns:
            tuple[np.ndarray, Any]:
                Tuple of a score tensor for y that has a shape `(len(next_tokens),)`
                and next state for ys
        """
        prev_score, state = state
        presub_score, new_st = self.impl(y, ids, state)
        tscore = np.array(
            presub_score - prev_score, dtype=x.dtype
        )
        return tscore, (presub_score, new_st)

    def batch_init_state(self, x: np.ndarray):
        """Get an initial state for decoding.
        Args:
            x (np.ndarray): The encoded feature tensor
        Returns: initial state
        """
        logp = self.ctc.run(["ctc_out"], {"x": x[None, :]})[0]
        xlen = np.array([logp.shape[1]])
        self.impl = CTCPrefixScoreTH(logp, xlen, 0, self.eos)
        return None

    def batch_score_partial(self, y, ids, state, x):
        """Score new token.
        Args:
            y (np.ndarray): 1D prefix token
            ids (np.ndarray): torch.int64 next token to score
            state: decoder state for prefix tokens
            x (np.ndarray): 2D encoder feature that generates ys
        Returns:
            tuple[np.ndarray, Any]:
                Tuple of a score tensor for y that has a shape `(len(next_tokens),)`
                and next state for ys
        """
        if state[0] is not None:
            batch_state = (
                np.concatenate([s[0][..., None] for s in state], axis=2),
                np.concatenate([s[1][None, :] for s in state]),
                state[0][2],
                state[0][3],
            )
        else:
            batch_state = None
        return self.impl(y, batch_state, ids)
    
    def extend_prob(self, x: np.ndarray):
        """Extend probs for decoding.

        This extension is for streaming decoding
        as in Eq (14) in https://arxiv.org/abs/2006.14941

        Args:
            x (np.ndarray): The encoded feature tensor

        """
        x = self.ctc.run(["ctc_out"], {"x": x[None, :]})[0]
        logp = log_softmax(x, axis=-1)
        self.impl.extend_prob(logp)

    def extend_state(self, state):
        """Extend state for decoding.

        This extension is for streaming decoding
        as in Eq (14) in https://arxiv.org/abs/2006.14941

        Args:
            state: The states of hyps

        Returns: exteded state

        """
        new_state = []
        for s in state:
            new_state.append(self.impl.extend_state(s))

        return new_state


class CTCPrefixScoreTH:
    """Batch processing of CTCPrefixScore
    which is based on Algorithm 2 in WATANABE et al.
    "HYBRID CTC/ATTENTION ARCHITECTURE FOR END-TO-END SPEECH RECOGNITION,"
    but extended to efficiently compute the label probablities for multiple
    hypotheses simultaneously
    See also Seki et al. "Vectorized Beam Search for CTC-Attention-Based
    Speech Recognition," In INTERSPEECH (pp. 3825-3829), 2019.
    """

    def __init__(self, x: np.ndarray, xlens: np.ndarray, blank: int, eos: int, margin: int = 0):
        """Construct CTC prefix scorer
        :param np.ndarray x: input label posterior sequences (B, T, O)
        :param np.ndarray xlens: input lengths (B,)
        :param int blank: blank label id
        :param int eos: end-of-sequence id
        :param int margin: margin parameter for windowing (0 means no windowing)
        """
        assert check_argument_types()
        # In the comment lines,
        # we assume T: input_length, B: batch size, W: beam width, O: output dim.
        self.logzero = -10000000000.0
        self.blank = blank
        self.eos = eos
        self.batch = x.shape[0]
        self.input_length = x.shape[1]
        self.odim = x.shape[2]
        self.dtype = x.dtype

        # Pad the rest of posteriors in the batch
        # TODO(takaaki-hori): need a better way without for-loops
        for i, l in enumerate(xlens):
            if l < self.input_length:
                x[i, l:, :] = self.logzero
                x[i, l:, blank] = 0
        # Reshape input x
        xn = x.transpose(1, 0, 2)  # (B, T, O) -> (T, B, O)
        xb = xn[:, :, None, self.blank].repeat(self.odim, axis=2)
        # operation is faster than np.stack
        self.x = np.concatenate([xn[None, :], xb[None, :]])
        self.end_frames = xlens - 1

        # Setup CTC windowing
        self.margin = margin
        if margin > 0:
            self.frame_ids = np.arange(
                self.input_length, dtype=self.dtype
            )
        # Base indices for index conversion
        self.idx_bh = None
        self.idx_b = np.arange(self.batch)
        self.idx_bo = (self.idx_b * self.odim)[:, None]

    def __call__(self, y, state, scoring_ids=None, att_w=None):
        """Compute CTC prefix scores for next labels
        :param list y: prefix label sequences
        :param tuple state: previous CTC state
        :param np.ndarray pre_scores: scores for pre-selection of hypotheses (BW, O)
        :param np.ndarray att_w: attention weights to decide CTC window
        :return new_state, ctc_local_scores (BW, O)
        """
        output_length = len(y[0]) - 1  # ignore blank and sos
        last_ids = [yi[-1] for yi in y]  # last output label ids
        n_bh = len(last_ids)  # batch * hyps
        n_hyps = n_bh // self.batch  # assuming each utterance has the same # of hyps
        self.scoring_num = scoring_ids.shape[-1] if scoring_ids is not None else 0
        # prepare state info
        if state is None:
            r_prev = np.full(
                (self.input_length, 2, self.batch, n_hyps),
                self.logzero,
                dtype=self.dtype,
            )
            r_prev[:, 1] = np.cumsum(self.x[0, :, :, self.blank], 0)[
                :, :, None]
            r_prev = r_prev.reshape(-1, 2, n_bh)
            s_prev = 0.0
            f_min_prev = 0
            f_max_prev = 1
        else:
            r_prev, s_prev, f_min_prev, f_max_prev = state

        # select input dimensions for scoring
        if self.scoring_num > 0:
            scoring_idmap = np.full(
                (n_bh, self.odim), -1, dtype=np.int64
            )
            snum = self.scoring_num
            if self.idx_bh is None or n_bh > len(self.idx_bh):
                self.idx_bh = np.arange(n_bh).reshape(-1, 1)
            scoring_idmap[self.idx_bh[:n_bh], scoring_ids] = np.arange(snum)
            scoring_idx = (
                scoring_ids + self.idx_bo.repeat(n_hyps, axis=1).reshape(-1, 1)
            ).reshape(-1)
            x_ = np.take(
                self.x.reshape(2, -1, self.batch * self.odim), scoring_idx, axis=2
            ).reshape(2, -1, n_bh, snum)
        else:
            scoring_ids = None
            scoring_idmap = None
            snum = self.odim
            x_ = self.x[:, :, :, None].repeat(
                n_hyps, axis=3).reshape(2, -1, n_bh, snum)

        # new CTC forward probs are prepared as a (T x 2 x BW x S) tensor
        # that corresponds to r_t^n(h) and r_t^b(h) in a batch.
        r = np.full(
            (self.input_length, 2, n_bh, snum),
            self.logzero,
            dtype=self.dtype
        )
        if output_length == 0:
            r[0, 0] = x_[0, 0]

        r_sum = logsumexp(r_prev, axis=1)
        log_phi = r_sum[:, :, None].repeat(snum, 2)
        if scoring_ids is not None:
            for idx in range(n_bh):
                pos = scoring_idmap[idx, int(last_ids[idx])]
                if pos >= 0:
                    log_phi[:, idx, pos] = r_prev[:, 1, idx]
        else:
            for idx in range(n_bh):
                log_phi[:, idx, last_ids[idx]] = r_prev[:, 1, idx]

        # decide start and end frames based on attention weights
        if att_w is not None and self.margin > 0:
            f_arg = np.matmul(att_w, self.frame_ids)
            f_min = max(int(f_arg.min()), f_min_prev)
            f_max = max(int(f_arg.max()), f_max_prev)
            start = min(f_max_prev, max(f_min - self.margin, output_length, 1))
            end = min(f_max + self.margin, self.input_length)
        else:
            f_min = f_max = 0
            start = max(output_length, 1)
            end = self.input_length

        # compute forward probabilities log(r_t^n(h)) and log(r_t^b(h))
        for t in range(start, end):
            rp = r[t - 1]
            rr = np.concatenate([rp[0:1], log_phi[t - 1:t], rp[0:1], rp[1:2]]).reshape(
                2, 2, n_bh, snum
            )
            r[t] = logsumexp(rr, 1) + x_[:, t]

        # compute log prefix probabilities log(psi)
        log_phi_x = np.concatenate(
            (log_phi[0][None, :], log_phi[:-1]), axis=0) + x_[0]
        if scoring_ids is not None:
            log_psi = np.full(
                (n_bh, self.odim), self.logzero, dtype=self.dtype
            )
            log_psi_ = logsumexp(
                np.concatenate(
                    (log_phi_x[start:end], r[start - 1, 0][None, :]), axis=0),
                axis=0,
            )
            for si in range(n_bh):
                log_psi[si, scoring_ids[si].astype(np.int64)] = log_psi_[si]
        else:
            log_psi = logsumexp(
                np.concatenate(
                    (log_phi_x[start:end], r[start - 1, 0][None, :]), axis=0),
                axis=0,
            )

        for si in range(n_bh):
            log_psi[si, self.eos] = r_sum[self.end_frames[si // n_hyps], si]

        # exclude blank probs
        log_psi[:, self.blank] = self.logzero

        return (log_psi - s_prev), (r, log_psi, f_min, f_max, scoring_idmap)

    def index_select_state(self, state, best_ids):
        """Select CTC states according to best ids
        :param state    : CTC state
        :param best_ids : index numbers selected by beam pruning (B, W)
        :return selected_state
        """
        r, s, f_min, f_max, scoring_idmap = state
        # convert ids to BHO space
        n_bh = len(s)
        n_hyps = n_bh // self.batch
        vidx = (best_ids + (self.idx_b * (n_hyps * self.odim)
                            ).reshape(-1, 1)).reshape(-1)
        # select hypothesis scores
        s_new = np.take(s.reshape(-1), vidx, axis=0)
        s_new = s_new.reshape(-1, 1).repeat(self.odim,
                                            axis=1).reshape(n_bh, self.odim)
        # convert ids to BHS space (S: scoring_num)
        if scoring_idmap is not None:
            snum = self.scoring_num
            hyp_idx = (best_ids // self.odim + (self.idx_b * n_hyps).reshape(-1, 1)).reshape(
                -1
            )
            label_ids = np.fmod(best_ids, self.odim).reshape(-1)
            score_idx = scoring_idmap[hyp_idx, label_ids]
            score_idx[score_idx == -1] = 0
            vidx = score_idx + hyp_idx * snum
        else:
            snum = self.odim
        # select forward probabilities
        r_new = np.take(r.reshape(-1, 2, n_bh * snum), vidx, axis=2).reshape(
            -1, 2, n_bh
        )
        return r_new, s_new, f_min, f_max

    def extend_prob(self, x):
        """Extend CTC prob.
        :param np.ndarray x: input label posterior sequences (B, T, O)
        """
        if self.x.shape[1] < x.shape[1]:  # self.x (2,T,B,O); x (B,T,O)
            # Pad the rest of posteriors in the batch
            # TODO(takaaki-hori): need a better way without for-loops
            xlens = np.array([x.shape[1]])
            for i, l in enumerate(xlens):
                if l < self.input_length:
                    x[i, l:, :] = self.logzero
                    x[i, l:, self.blank] = 0
            tmp_x = self.x
            xn = x.transpose(1, 0, 2)  # (B, T, O) -> (T, B, O)
            xb = xn[:, :, None, self.blank].repeat(self.odim, axis=2)
            self.x = np.concatenate([xn[None, :], xb[None, :]])  # (2, T, B, O)
            self.x[:, : tmp_x.shape[1], :, :] = tmp_x
            self.input_length = x.shape[1]
            self.end_frames = xlens - 1

    def extend_state(self, state):
        """Compute CTC prefix state.
        :param state    : CTC state
        :return ctc_state
        """

        if state is None:
            # nothing to do
            return state
        else:
            r_prev, s_prev, f_min_prev, f_max_prev = state

            r_prev_new = np.full(
                (self.input_length, 2),
                self.logzero,
                dtype=self.dtype,
            )
            start = max(r_prev.shape[0], 1)
            r_prev_new[0:start] = r_prev
            for t in six.moves.range(start, self.input_length):
                r_prev_new[t, 1] = r_prev_new[t - 1, 1] + \
                    self.x[0, t, :, self.blank]

            return (r_prev_new, s_prev, f_min_prev, f_max_prev)



from typing import List
from typing import Any
from typing import Tuple
from typing import Optional
from typing import Dict
from typing import Union

import numpy as np
import onnxruntime

from espnet_onnx.asr.scorer.interface import BatchScorerInterface
from espnet_onnx.asr.beam_search.hyps import TransducerHypothesis
from espnet_onnx.utils.function import subsequent_mask
from espnet_onnx.utils.config import Config


class XformerDecoder(BatchScorerInterface):
    def __init__(
        self,
        config: Config,
        providers: List[str],
        use_quantized: bool = False
    ):
        """Onnx support for espnet2.asr.decoder.transformer_decoder

        Args:
            config (Config):
            use_quantized (bool): Flag to use quantized model
        """
        if use_quantized:
            self.decoder = onnxruntime.InferenceSession(
                config.quantized_model_path,
                providers=providers
            )
        else:
            self.decoder = onnxruntime.InferenceSession(
                config.model_path,
                providers=providers
            )
        self.n_layers = config.n_layers
        self.odim = config.odim
        self.in_caches = [d.name for d in self.decoder.get_inputs()
                          if 'cache' in d.name]
        self.out_caches = [d.name for d in self.decoder.get_outputs()
                           if 'cache' in d.name]

    def batch_score(
        self, ys: np.ndarray, states: List[Any], xs: np.ndarray
    ) -> Tuple[np.ndarray, List[Any]]:
        """Score new token batch.
        Args:
            ys (np.ndarray): np.int64 prefix tokens (n_batch, ylen).
            states (List[Any]): Scorer states for prefix tokens.
            xs (np.ndarray):
                The encoder feature that generates ys (n_batch, xlen, n_feat).
        Returns:
            tuple[np.ndarray, List[Any]]: Tuple of
                batchfied scores for next token with shape of `(n_batch, n_vocab)`
                and next state list for ys.
        """
        # merge states
        if len(ys.shape) == 1:
            ys = ys[None, :]

        n_batch = len(ys)
        if states[0] is None:
            batch_state = [
                np.zeros((1, 1, self.odim), dtype=np.float32)
                for _ in range(self.n_layers)
            ]
        else:
            # transpose state of [batch, layer] into [layer, batch]
            batch_state = [
                np.concatenate([states[b][i][None, :] for b in range(n_batch)])
                for i in range(self.n_layers)
            ]

        # batch decoding
        ys_mask = subsequent_mask(ys.shape[-1])[None, :]
        input_dict = self.get_input_dict(ys, ys_mask, xs, batch_state)

        logp, *states = self.decoder.run(
            ['y'] + self.out_caches,
            input_dict
        )

        # transpose state of [layer, batch] into [batch, layer]
        state_list = [[states[i][b]
                       for i in range(self.n_layers)] for b in range(n_batch)]
        return logp, state_list

    def get_input_dict(self, ys, ys_mask, xs, state):
        in_names = [d.name for d in self.decoder.get_inputs() if 'cache' not in d.name]
        ret = {}
        if 'tgt' in in_names: ret.update(tgt=ys.astype(np.int64))
        if 'tgt_mask' in in_names: ret.update(tgt_mask=ys_mask)
        if 'memory' in in_names: ret.update(memory=xs)
        ret.update(
            {k: v for (k, v) in zip(self.in_caches, state)})
        return ret
        


from typing import List
from typing import Dict

import numpy as np
import onnxruntime

from espnet_onnx.asr.scorer.interface import BatchScorerInterface
from espnet_onnx.utils.function import make_pad_mask
from espnet_onnx.utils.config import Config


class RNNDecoder(BatchScorerInterface):
    def __init__(
        self,
        config: Config,
        providers: List[str],
        use_quantized: bool = False
    ):
        """Onnx support for espnet2.asr.decoder.rnn_decoder.RNNDecoder

        Args:
            config (Config):
            use_quantized (bool): Flag to use quantized model
        """
        self.config = config
        # predecoder
        self.predecoders = []
        for p in config.predecoder:
            if use_quantized:
                model_path = p.quantized_model_path
            else:
                model_path = p.model_path
            self.predecoders.append(
                onnxruntime.InferenceSession(
                    model_path,
                    providers=providers
                )
            )

        # decoder
        if use_quantized:
            self.decoder = onnxruntime.InferenceSession(
                config.quantized_model_path,
                providers=providers
            )
        else:
            self.decoder = onnxruntime.InferenceSession(
                config.model_path,
                providers=providers
            )
            
        # HP
        self.num_encs = len(self.predecoders)
        self.dunits = config.dunits
        self.dlayers = config.dlayers
        self.rnn_type = config.rnn_type
        self.decoder_length = config.decoder_length

        # predecoder
        self.decoder_output_names = self.get_decoder_output_names()

        # cache pre_computed features
        self.pre_compute_enc_h = []
        self.enc_h = []
        self.mask = []
        
        self.init_input_names()
        
    def init_input_names(self):
        # a_prev, enc_h, pceh, mask is not required
        # with some attention type.
        self.required_input_names = {}
        input_names = [d.name for d in self.decoder.get_inputs()]
        self.required_input_names['a_prev'] = True \
            if "a_prev" in str(input_names) else False
        self.required_input_names['enc_h'] = True \
            if "enc_h" in str(input_names) else False
        self.required_input_names['pceh'] = True \
            if "pceh" in str(input_names) else False
        self.required_input_names['mask'] = True \
            if "mask" in str(input_names) else False

    def get_decoder_output_names(self):
        return [d.name for d in self.decoder.get_outputs()]

    def zero_state(self, hs_pad):
        return np.zeros((hs_pad.shape[0], self.dunits), dtype=np.float32)

    def get_att_prev(self, x, att_type=None):
        att_prev = 1.0 - make_pad_mask([x[0].shape[0]])
        att_prev = (
            att_prev / np.array([x[0].shape[0]])[..., None]).astype(np.float32)
        if att_type == 'location2d':
            att_prev = att_prev[..., None].reshape(-1, self.config.att_win, -1)
        if att_type in ('coverage', 'coverage_location'):
            att_prev = att_prev[:, None, :]
        return att_prev

    def init_state(self, x):
        # to support mutiple encoder asr mode, in single encoder mode,
        # convert torch.Tensor to List of torch.Tensor
        if self.num_encs <= 1:
            x = [x]

        c_list = [self.zero_state(x[0][None, :])]
        z_list = [self.zero_state(x[0][None, :])]
        for _ in range(1, self.dlayers):
            c_list.append(self.zero_state(x[0][None, :]))
            z_list.append(self.zero_state(x[0][None, :]))

        strm_index = 0
        att_idx = min(strm_index, len(self.predecoders) - 1)

        a = []
        for att in self.config.predecoder:
            a += [self.get_att_prev(x, att.att_type)]
        
        if self.num_encs != 1:
            a += [self.get_att_prev(x)]  # atts + han

        # initialize cached parameters
        self.pre_compute_enc_h = []
        self.enc_h = []
        self.mask = []
        return dict(
            c_prev=c_list[:],
            z_prev=z_list[:],
            a_prev=a,
            workspace=(att_idx, z_list, c_list),
        )

    def score(self, yseq, state, x):
        att_idx, z_list, c_list = state["workspace"]
        vy = np.array([yseq[-1]])

        if self.num_encs <= 1:
            x = [x]
        
        # set initial state if attention type is NoAtt
        if self.num_encs == 0:
            # we d
            self.enc_h.append(x[0][None, :])
            self.mask.append(np.where(make_pad_mask(
                [x[0].shape[0]]) == 1, -float('inf'), 0).astype(np.float32))

        # pre compute states of attention.
        if len(self.pre_compute_enc_h) == 0:
            for idx in range(self.num_encs):
                _pceh = self.predecoders[idx].run(
                    ['pre_compute_enc_h'], {
                        'enc_h': x[idx][None, :]
                    }
                )[0]
                self.pre_compute_enc_h.append(_pceh)
                self.enc_h.append(x[idx][None, :])
                self.mask.append(np.where(make_pad_mask(
                    [x[idx].shape[0]]) == 1, -float('inf'), 0).astype(np.float32))

        input_dict = self.create_input_dic(vy, x, state)
        logp, *status_lists = self.decoder.run(
            self.decoder_output_names,
            input_dict
        )
        c_list, z_list, att_w = self._split(status_lists)
        return (
            logp,
            dict(
                c_prev=c_list,
                z_prev=z_list,
                a_prev=att_w,
                workspace=(att_idx, z_list, c_list),
            ),
        )

    def create_input_dic(self, vy, x, state):
        ret = {
            'vy': vy.astype(np.int64),
        }
        ret.update({
            'z_prev_%d' % d: state['z_prev'][d]
            for d in range(self.decoder_length)
        })
        ret.update({
            'c_prev_%d' % d: state['c_prev'][d]
            for d in range(self.decoder_length)
        })
        necs = max(1, self.num_encs)
        if self.required_input_names['a_prev']:
            ret.update({
                'a_prev_%d' % d: state['a_prev'][d]
                for d in range(necs)
            })
        if self.required_input_names['enc_h']:
            ret.update({
                'enc_h_%d' % d: self.enc_h[d]
                for d in range(necs)
            })
        if self.required_input_names['pceh']:
            ret.update({
                'pceh_%d' % d: self.pre_compute_enc_h[d]
                for d in range(self.num_encs)
            })
        if self.required_input_names['mask']:
            ret.update({
                'mask_%d' % d: self.mask[d]
                for d in range(self.num_encs)
            })
        return ret

    def _split(self, status_lists):
        c_list = status_lists[:self.decoder_length]
        z_list = status_lists[self.decoder_length: 2*self.num_encs]
        att_w = status_lists[2*self.decoder_length:]
        return c_list, z_list, att_w



from typing import (
    Optional,
    Tuple,
    List,
    Dict,
)

import onnxruntime
import numpy as np
from espnet_onnx.utils.config import Config


class Vocoder:
    def __init__(
        self,
        config: Config,
        providers: List[str],
        use_quantized: bool = False,
    ):
        self.config = config
        if use_quantized:
            self.model = onnxruntime.InferenceSession(
                self.config.quantized_model_path,
                providers=providers
            )
        else:
            self.model = onnxruntime.InferenceSession(
                self.config.model_path,
                providers=providers
            )

    def __call__(
        self,
        c: np.ndarray
    ):
        input_dict = {'c': c}
        return self.model.run(['wav'], input_dict)[0]




from typing import (
    Optional,
    Tuple
)

import math
import torch
import torch.nn as nn

from espnet_onnx.export.asr.models.language_models.embed import get_pos_emb
from espnet_onnx.utils.torch_function import MakePadMask
from espnet_onnx.utils.abs_model import AbsExportModel


class OnnxTextEncoder(nn.Module):
    def __init__(self, model, make_pad_mask, max_seq_len):
        super().__init__()
        self.model = model
        # fix RelPositionalEncoding
        self.model.encoder.embed[0] = get_pos_emb(self.model.encoder.embed[0], max_seq_len)
        self.make_pad_mask = make_pad_mask

    def forward(self, x, x_lengths):
        x = self.model.emb(x) * math.sqrt(self.model.attention_dim)
        x_mask = 1 - self.make_pad_mask(x_lengths).unsqueeze(1)
        # encoder assume the channel last (B, T_text, attention_dim)
        # but mask shape shoud be (B, 1, T_text)
        x, _ = self.model.encoder(x, x_mask)

        # convert the channel first (B, attention_dim, T_text)
        x = x.transpose(1, 2)
        stats = self.model.proj(x) * x_mask
        m, logs = stats.split(stats.size(1) // 2, dim=1)
        return x, m, logs, x_mask


class OnnxVITSGenerator(nn.Module):
    def __init__(
        self,
        model,
        max_seq_len: int = None,
        noise_scale: float = 0.667,
        noise_scale_dur: float = 0.8,
        alpha: float = 1.0,
        use_teacher_forcing: bool = False
    ):
        super().__init__()
        self.make_pad_mask = MakePadMask(max_seq_len)
        self.text_encoder = OnnxTextEncoder(model.text_encoder, self.make_pad_mask, max_seq_len)
        self.decoder = model.decoder
        self.posterior_encoder = model.posterior_encoder
        self.flow = model.flow
        self.duration_predictor = model.duration_predictor
        self.model = model
        self.noise_scale = noise_scale
        self.noise_scale_dur = noise_scale_dur
        self.alpha = alpha
        self.use_teacher_forcing = use_teacher_forcing
        self.max_seq_len = max_seq_len

    def forward(
        self,
        text: torch.Tensor,
        text_lengths: torch.Tensor,
        feats: Optional[torch.Tensor] = None,
        feats_lengths: Optional[torch.Tensor] = None,
        sids: Optional[torch.Tensor] = None,
        spembs: Optional[torch.Tensor] = None,
        lids: Optional[torch.Tensor] = None,
        dur: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Run inference.

        Args:
            text (Tensor): Input text index tensor (B, T_text,).
            text_lengths (Tensor): Text length tensor (B,).
            feats (Tensor): Feature tensor (B, aux_channels, T_feats,).
            feats_lengths (Tensor): Feature length tensor (B,).
            sids (Optional[Tensor]): Speaker index tensor (B,) or (B, 1).
            spembs (Optional[Tensor]): Speaker embedding tensor (B, spk_embed_dim).
            lids (Optional[Tensor]): Language index tensor (B,) or (B, 1).
            dur (Optional[Tensor]): Ground-truth duration (B, T_text,). If provided,
                skip the prediction of durations (i.e., teacher forcing).
            noise_scale (float): Noise scale parameter for flow.
            noise_scale_dur (float): Noise scale parameter for duration predictor.
            alpha (float): Alpha parameter to control the speed of generated speech.
            max_len (Optional[int]): Maximum length of acoustic feature sequence.
            use_teacher_forcing (bool): Whether to use teacher forcing.

        Returns:
            Tensor: Generated waveform tensor (B, T_wav).
            Tensor: Monotonic attention weight tensor (B, T_feats, T_text).
            Tensor: Duration tensor (B, T_text).

        """
        # encoder
        x, m_p, logs_p, x_mask = self.text_encoder(text, text_lengths)
        g = None
        if self.model.spks is not None:
            # (B, global_channels, 1)
            g = self.model.global_emb(sids.view(-1)).unsqueeze(-1)
        if self.model.spk_embed_dim is not None:
            # (B, global_channels, 1)
            g_ = self.model.spemb_proj(F.normalize(
                spembs.unsqueeze(0))).unsqueeze(-1)
            if g is None:
                g = g_
            else:
                g = g + g_
        if self.model.langs is not None:
            # (B, global_channels, 1)
            g_ = self.model.lang_emb(lids.view(-1)).unsqueeze(-1)
            if g is None:
                g = g_
            else:
                g = g + g_

        if self.use_teacher_forcing:
            # forward posterior encoder
            z, m_q, logs_q, y_mask = self.posterior_encoder(
                feats, feats_lengths, g=g)

            # forward flow
            z_p = self.flow(z, y_mask, g=g)  # (B, H, T_feats)

            # monotonic alignment search
            s_p_sq_r = torch.exp(-2 * logs_p)  # (B, H, T_text)
            # (B, 1, T_text)
            neg_x_ent_1 = torch.sum(
                -0.5 * math.log(2 * math.pi) - logs_p,
                [1],
                keepdim=True,
            )
            # (B, T_feats, H) x (B, H, T_text) = (B, T_feats, T_text)
            neg_x_ent_2 = torch.matmul(
                -0.5 * (z_p**2).transpose(1, 2),
                s_p_sq_r,
            )
            # (B, T_feats, H) x (B, H, T_text) = (B, T_feats, T_text)
            neg_x_ent_3 = torch.matmul(
                z_p.transpose(1, 2),
                (m_p * s_p_sq_r),
            )
            # (B, 1, T_text)
            neg_x_ent_4 = torch.sum(
                -0.5 * (m_p**2) * s_p_sq_r,
                [1],
                keepdim=True,
            )
            # (B, T_feats, T_text)
            neg_x_ent = neg_x_ent_1 + neg_x_ent_2 + neg_x_ent_3 + neg_x_ent_4
            # (B, 1, T_feats, T_text)
            attn_mask = torch.unsqueeze(
                x_mask, 2) * torch.unsqueeze(y_mask, -1)
            # monotonic attention weight: (B, 1, T_feats, T_text)
            attn = self.model.maximum_path(
                neg_x_ent,
                attn_mask.squeeze(1),
            ).unsqueeze(1)
            dur = attn.sum(2)  # (B, 1, T_text)

            # forward decoder with random segments
            wav = self.decoder(z * y_mask, g=g)
        else:
            # duration
            if dur is None:
                logw = self.duration_predictor(
                    x,
                    x_mask,
                    g=g,
                    inverse=True,
                    noise_scale=self.noise_scale_dur,
                )
                w = torch.exp(logw) * x_mask * self.alpha
                dur = torch.ceil(w)
            y_lengths = torch.clamp_min(torch.sum(dur, [1, 2]), 1).long()
            y_mask = 1 - self.make_pad_mask(y_lengths).unsqueeze(1)
            attn_mask = torch.unsqueeze(
                x_mask, 2) * torch.unsqueeze(y_mask, -1)
            attn = self.model._generate_path(dur, attn_mask)

            # expand the length to match with the feature sequence
            # (B, T_feats, T_text) x (B, T_text, H) -> (B, H, T_feats)
            m_p = torch.matmul(
                attn.squeeze(1),
                m_p.transpose(1, 2),
            ).transpose(1, 2)
            # (B, T_feats, T_text) x (B, T_text, H) -> (B, H, T_feats)
            logs_p = torch.matmul(
                attn.squeeze(1),
                logs_p.transpose(1, 2),
            ).transpose(1, 2)

            # decoder
            z_p = m_p + torch.randn_like(m_p) * torch.exp(logs_p) * self.noise_scale
            z = self.flow(z_p, y_mask, g=g, inverse=True)
            wav = self.decoder((z * y_mask)[:, :, :self.max_seq_len], g=g)

        return wav.squeeze(1), attn.squeeze(1), dur.squeeze(1)


class OnnxVITSModel(nn.Module, AbsExportModel):
    def __init__(
        self,
        model,
        noise_scale: float = 0.667,
        noise_scale_dur: float = 0.8,
        alpha: float = 1.0,
        max_seq_len: int = 512,
        use_teacher_forcing: bool = False,
        predict_duration: bool = True,
        **kwargs
    ):
        super().__init__()
        self.model = model
        self.generator = OnnxVITSGenerator(
            model.generator,
            max_seq_len,
            noise_scale,
            noise_scale_dur,
            alpha,
            use_teacher_forcing
        )
        self.use_teacher_forcing = use_teacher_forcing
        self.predict_duration = predict_duration
        self.model_name = 'vits'

    def forward(
        self,
        text: torch.Tensor,
        text_lengths: torch.Tensor,
        feats: Optional[torch.Tensor] = None,
        feats_lengths: Optional[torch.Tensor] = None,
        sids: Optional[torch.Tensor] = None,
        spembs: Optional[torch.Tensor] = None,
        lids: Optional[torch.Tensor] = None,
        durations: Optional[torch.Tensor] = None,
    ):
        # setup
        text = text[None]

        if self.use_teacher_forcing:
            assert feats is not None
            feats = feats[None].transpose(1, 2)
            wav, att_w, dur = self.generator(
                text=text,
                text_lengths=text_lengths,
                feats=feats,
                feats_lengths=feats_lengths,
                sids=sids,
                spembs=spembs,
                lids=lids,
            )
        else:
            wav, att_w, dur = self.generator(
                text=text,
                text_lengths=text_lengths,
                sids=sids,
                spembs=spembs,
                lids=lids,
                dur=durations,
            )
        return dict(wav=wav.view(-1), att_w=att_w[0], duration=dur[0])

    def get_dummy_inputs(self):
        text = torch.LongTensor([0, 1])
        text_length = torch.LongTensor([text.size(0)])
        feats = torch.randn(5, self.model.generator.posterior_encoder.input_conv.in_channels) \
            if self.use_teacher_forcing else None
        feats_length = torch.LongTensor([5]) if feats is not None else None

        sids = torch.LongTensor([0]) \
            if self.model.generator.spks is not None else None

        spembs = torch.randn(self.model.generator.spk_embed_dim) \
            if self.model.generator.spks is not None else None

        lids = torch.LongTensor([0]) \
            if self.model.generator.langs is not None else None

        duration = torch.randn(text.size(0)) \
            if not self.predict_duration else None

        return (text, text_length, feats, feats_length, sids, spembs, lids, duration)

    def get_input_names(self):
        return ['text', 'text_length', 'feats', 'feats_length', 'sids', 'spembs', 'lids', 'duration']

    def get_output_names(self):
        return ['wav', 'att_w', 'dur']

    def get_dynamic_axes(self):
        return {
            'text': {0: 'text_length'},
            'feats': {0: 'feats_length'},
            'duration': {0: 'duration_length'},
            'wav': {0: 'wav_length'},
            'att_w': {0: 'att_w_feat_length',
                      1: 'att_w_text_length'},
            'dur': {0: 'dur_length'}
        }

    def get_model_config(self, path):
        return {
            'model_type': 'VITS',
            'model_path': str(path / f'{self.model_name}.onnx')
        }



from espnet2.gan_tts.vits import VITS as espnetVITSModel
from espnet2.tts.fastspeech2.fastspeech2 import FastSpeech2 as espnetFastSpeech2
from espnet2.tts.tacotron2.tacotron2 import Tacotron2 as espnetTacotron2
from espnet2.gan_tts.joint.joint_text2wav import JointText2Wav as espnetJointText2Wav
from .tts_models.vits import OnnxVITSModel
from .tts_models.fastspeech2 import OnnxFastSpeech2
from .tts_models.tacotron2 import (
    OnnxTacotron2Encoder,
    OnnxTacotron2Decoder
)

# vocoder
from espnet2.utils.griffin_lim import Spectrogram2Waveform
from espnet2.tts.utils.parallel_wavegan_pretrained_vocoder import ParallelWaveGANPretrainedVocoder
from espnet2.gan_tts.hifigan.hifigan import HiFiGANGenerator
from espnet2.gan_tts.melgan.melgan import MelGANGenerator
from espnet2.gan_tts.parallel_wavegan.parallel_wavegan import ParallelWaveGANGenerator
from espnet2.gan_tts.style_melgan.style_melgan import StyleMelGANGenerator
from .vocoders.hifigan import OnnxHiFiGANVocoder
from .vocoders.melgan import OnnxMelGANVocoder
from .vocoders.parallel_wavegan import OnnxPWGVocoder
from .vocoders.style_melgan import OnnxStyleMelGANVocoder


def get_tts_model(model, export_config):
    if isinstance(model, espnetVITSModel):
        return OnnxVITSModel(model, **export_config)
    
    elif isinstance(model, espnetFastSpeech2):
        return OnnxFastSpeech2(model, **export_config)
    
    elif isinstance(model, espnetTacotron2):
        return [
            OnnxTacotron2Encoder(model, **export_config),
            OnnxTacotron2Decoder(model, **export_config),
            'Tacotron2'
        ]   
    
    elif isinstance(model, espnetJointText2Wav):
        return get_tts_model(model.generator['text2mel'], export_config)
    else:
        raise RuntimeError('Currently, VITS and FastSpeech2 is supported.')

def get_vocoder(model, export_config):
    if isinstance(model, HiFiGANGenerator):
        return OnnxHiFiGANVocoder(model, **export_config), True
    elif isinstance(model, MelGANGenerator):
        return OnnxMelGANVocoder(model, **export_config), True
    elif isinstance(model, ParallelWaveGANGenerator):
        return OnnxPWGVocoder(model, **export_config), True
    elif isinstance(model, StyleMelGANGenerator):
        # return OnnxStyleMelGANVocoder(model, **export_config), True
        raise RuntimeError('Currently, StyleMelgan is not supported.')
    
    if hasattr(model, 'vocoder'):
        if isinstance(model.vocoder, Spectrogram2Waveform):
            return model.vocoder, False
        elif isinstance(model.vocoder, ParallelWaveGANPretrainedVocoder):
            raise RuntimeError('Pretrained PWGVocoder is currently not supported.')
    
    if hasattr(model, 'model'):
        if isinstance(model.model.tts, espnetJointText2Wav):
            return get_vocoder(model.model.tts.generator['vocoder'], export_config)



from typing import Union
from typing import Tuple
from typing import Any
from typing import List

import numpy as np
import onnxruntime
from scipy.special import log_softmax

from espnet_onnx.asr.scorer.interface import BatchScorerInterface
from espnet_onnx.utils.function import subsequent_mask


class TransformerLM(BatchScorerInterface):
    def __init__(
        self,
        config,
        providers: List[str],
        use_quantized=False
    ):
        if use_quantized:
            self.lm_session = onnxruntime.InferenceSession(
                config.quantized_model_path,
                providers=providers
            )
        else:
            self.lm_session = onnxruntime.InferenceSession(
                config.model_path,
                providers=providers
            )
        self.enc_output_names = ['y'] \
            + [d.name for d in self.lm_session.get_outputs() if 'cache' in d.name]
        self.enc_in_cache_names = [
            d.name for d in self.lm_session.get_inputs() if 'cache' in d.name]

        self.nlayers = config.nlayers
        self.odim = config.odim

    def _target_mask(self, ys_in_pad):
        # ys_in_pad : (B, D)
        ys_mask = ys_in_pad != 0
        m = subsequent_mask(ys_mask.shape[-1])[None, :]
        return ys_mask[:, None, :] * m

    def score(
        self, y: np.ndarray, state: Any, x: np.ndarray
    ) -> Tuple[np.ndarray, Any]:
        """Score new token.

        Args:
            y (np.ndarray): 1D torch.int64 prefix tokens.
            state: Scorer state for prefix tokens
            x (np.ndarray): encoder feature that generates ys.

        Returns:
            tuple[np.ndarray, Any]: Tuple of
                torch.float32 scores for next token (vocab_size)
                and next state for ys

        """
        y = y[None, :]

        input_dic = {'tgt': y, 'tgt_mask': self._target_mask(y)}
        input_dic.update({
            k: v for k, v in zip(self.enc_in_cache_names, state)
        })
        decoded, *new_state = self.lm_session.run(
            self.enc_output_names,
            input_dic
        )

        logp = log_softmax(decoded, axis=-1).squeeze(0)
        return logp, new_state

    def batch_score(
        self, ys: np.ndarray, states: List[Any], xs: np.ndarray
    ) -> Tuple[np.ndarray, List[Any]]:
        """Score new token batch.

        Args:
            ys (np.ndarray): torch.int64 prefix tokens (n_batch, ylen).
            states (List[Any]): Scorer states for prefix tokens.
            xs (np.ndarray):
                The encoder feature that generates ys (n_batch, xlen, n_feat).

        Returns:
            tuple[np.ndarray, List[Any]]: Tuple of
                batchfied scores for next token with shape of `(n_batch, vocab_size)`
                and next state list for ys.

        """
        # merge states
        ys = ys.astype(np.int64)
        n_batch = len(ys)
        if states[0] is None:
            batch_state = [
                np.zeros((1, 1, self.odim), dtype=np.float32)
                for _ in range(self.nlayers)
            ]
        else:
            # transpose state of [batch, layer] into [layer, batch]
            batch_state = [
                np.concatenate([states[b][i][None, :] for b in range(n_batch)])
                for i in range(self.nlayers)
            ]

        input_dic = {'tgt': ys, 'tgt_mask': self._target_mask(ys)}
        input_dic.update({
            k: v for k, v in zip(self.enc_in_cache_names, batch_state)
        })
        decoded, *new_state = self.lm_session.run(
            self.enc_output_names,
            input_dic
        )
        logp = log_softmax(decoded, axis=-1)

        # transpose state of [layer, batch] into [batch, layer]
        state_list = [[new_state[i][b]
                       for i in range(self.nlayers)] for b in range(n_batch)]
        return logp, state_list



from typing import Union
from pathlib import Path
from typeguard import check_argument_types

import numpy as np
import onnxruntime

from .stft import Stft
from .logmel import LogMel

from espnet_onnx.utils.config import Config


class Frontend:
    """Default frontend module.
    This class is based on espnet2.asr.frontend.default.DefaultFrontend

    Args:
        config (Config): configuration for default frontend
    """

    def __init__(
        self,
        config: Config,
    ):
        self.stft = Stft(config.stft)
        self.logmel = LogMel(config.logmel)

    def __call__(self, inputs: np.ndarray, input_length: np.ndarray):
        assert check_argument_types()
        # STFT
        input_stft, feats_lens = self.stft(inputs, input_length)

        # 3. STFT -> Power spectrum
        # h: ComplexTensor(B, T, F) -> torch.Tensor(B, T, F)
        input_power = input_stft[..., 0]**2 + input_stft[..., 1]**2

        # 4. Feature transform e.g. Stft -> Log-Mel-Fbank
        # input_power: (Batch, [Channel,] Length, Freq)
        #       -> input_feats: (Batch, Length, Dim)
        input_feats, _ = self.logmel(input_power, feats_lens)
        return input_feats, feats_lens



from typing import Union
from typing import List
from typing import Tuple
from typing import Optional
from pathlib import Path
from typeguard import check_argument_types
import warnings

import os
import logging
import numpy as np
import onnxruntime

from espnet_onnx.asr.abs_asr_model import AbsASRModel
from espnet_onnx.asr.beam_search.hyps import Hypothesis



class Speech2Text(AbsASRModel):
    """Wrapper class for espnet2.asr.bin.asr_infer.Speech2Text

    """

    def __init__(self,
                 tag_name: str = None,
                 model_dir: Union[Path, str] = None,
                 providers: List[str] = ['CPUExecutionProvider'],
                 use_quantized: bool = False,
                 ):
        assert check_argument_types()
        self._check_argument(tag_name, model_dir)
        self._load_config()
        
        # check onnxruntime version and providers
        self._check_ort_version(providers)
        
        # check if model is exported for streaming.
        if self.config.encoder.enc_type == 'ContextualXformerEncoder':
            raise RuntimeError('Onnx model is built for streaming. Use StreamingSpeech2Text instead.')

        if use_quantized and 'quantized_model_path' not in self.config.encoder.keys():
            # check if quantized model config is defined.
            raise RuntimeError(
                'Configuration for quantized model is not defined.')

        self._build_model(providers, use_quantized)
        
        if self.config.transducer.use_transducer_decoder:
            self.start_idx = 1
            self.last_idx = None
        else:
            self.start_idx = 2
            self.last_idx = -1


    def __call__(self, speech: np.ndarray) -> List[
        Tuple[
            Optional[str],
            List[str],
            List[int],
            Union[Hypothesis],
        ]
    ]:
        """Inference
        Args:
            data: Input speech data
        Returns:
            text, token, token_int, hyp
        """
        assert check_argument_types()

        # check dtype
        if speech.dtype != np.float32:
            speech = speech.astype(np.float32)

        # data: (Nsamples,) -> (1, Nsamples)
        speech = speech[np.newaxis, :]
        # lengths: (1,)
        lengths = np.array([speech.shape[1]]).astype(np.int64)

        # b. Forward Encoder
        enc, _ = self.encoder(speech=speech, speech_length=lengths)
        if isinstance(enc, tuple):
            enc = enc[0]
        assert len(enc) == 1, len(enc)

        nbest_hyps = self.beam_search(enc[0])[:1]

        results = []
        for hyp in nbest_hyps:
            # remove sos/eos and get results
            if self.last_idx is not None:
                token_int = list(hyp.yseq[self.start_idx : self.last_idx])
            else:
                token_int = list(hyp.yseq[self.start_idx:])
                
            # remove blank symbol id, which is assumed to be 0
            token_int = list([int(i) for i in filter(lambda x: x != 0, token_int)])

            # Change integer-ids to tokens
            token = self.converter.ids2tokens(token_int)

            if self.tokenizer is not None:
                text = self.tokenizer.tokens2text(token)
            else:
                text = None
            results.append((text, token, token_int, hyp))

        return results



from abc import ABC

from typing import List
import torch.nn as nn

import os
import glob
import logging
import onnxruntime
import warnings

from espnet_onnx.utils.config import (
    get_config,
    get_tag_config
)
from espnet_onnx.asr.postprocess.build_tokenizer import build_tokenizer
from espnet_onnx.asr.postprocess.token_id_converter import TokenIDConverter


class AbsModel(ABC):

    def _check_argument(self, tag_name, model_dir):
        self.model_dir = model_dir

        if tag_name is None and model_dir is None:
            raise ValueError('tag_name or model_dir should be defined.')

        if tag_name is not None:
            tag_config = get_tag_config()
            if tag_name not in tag_config.keys():
                raise RuntimeError(f'Model path for tag_name "{tag_name}" is not set on tag_config.yaml.'
                                   + 'You have to export to onnx format with `espnet_onnx.export.asr.export_asr.ModelExport`,'
                                   + 'or have to set exported model path in tag_config.yaml.')
            self.model_dir = tag_config[tag_name]

    def _load_config(self):
        config_file = glob.glob(os.path.join(self.model_dir, 'config.*'))[0]
        self.config = get_config(config_file)

    def _build_tokenizer(self):
        if self.config.tokenizer.token_type is None:
            self.tokenizer = None
        elif self.config.tokenizer.token_type == 'bpe':
            self.tokenizer = build_tokenizer(
                'bpe', self.config.tokenizer.bpemodel)
        else:
            self.tokenizer = build_tokenizer(
                **self.config.tokenizer.dic)

    def _build_token_converter(self):
        self.converter = TokenIDConverter(token_list=self.config.token.list)

    def _build_model(self, providers, use_quantized):
        raise NotImplementedError

    def _check_ort_version(self, providers: List[str]):
        # check cpu
        if onnxruntime.get_device() == 'CPU' and 'CPUExecutionProvider' not in providers:
            raise RuntimeError(
                'If you want to use GPU, then follow `How to use GPU on espnet_onnx` chapter in readme to install onnxruntime-gpu.')

        # check GPU
        if onnxruntime.get_device() == 'GPU' and providers == ['CPUExecutionProvider']:
            warnings.warn(
                'Inference will be executed on the CPU. Please provide gpu providers. Read `How to use GPU on espnet_onnx` in readme in detail.')

        logging.info(f'Providers [{" ,".join(providers)}] detected.')


class AbsExportModel(ABC):
    def forward(self, *args, **kwargs):
        raise NotImplementedError

    def get_input_names(self):
        raise NotImplementedError

    def get_output_names(self):
        raise NotImplementedError

    def get_dynamix_axes(self):
        return {}

    def get_model_config(self):
        raise NotImplementedError



from typing import (
    Optional,
    Tuple
)

import math
import torch
import torch.nn as nn
import torch.nn.functional as F

from espnet_onnx.export.asr.models.language_models.embed import Embedding
from espnet_onnx.utils.torch_function import MakePadMask
from espnet_onnx.utils.abs_model import AbsExportModel


class OnnxLengthRegurator(nn.Module):
    def __init__(self, alpha):
        super().__init__()
        self.alpha = alpha
        # The maximum length of the make_pad_mask is the 
        # maximum value of the duration.
        self.make_pad_mask = MakePadMask(512)
        self.zero = torch.LongTensor([0])
    
    def forward(self, x, dur):
        # This class assumes that the batch size of x
        # should be 1.
        if self.alpha != 1.0:
            dur = torch.round(ds.float() * alpha).long()
        dur = dur[0]
        ma = torch.cumsum(
            torch.cat((dur, self.zero), dim=0),
            dim=0
        )
        mb = torch.cumsum(
            torch.cat((self.zero, dur), dim=0),
            dim=0
        )
        _m = self.make_pad_mask(mb)
        _m[0] = 1
        mask = ((1 - self.make_pad_mask(ma)) * _m)[:-1]
        return torch.matmul(x.transpose(1, 2), mask).transpose(1, 2)


class OnnxReferenceEncoder(nn.Module):
    def __init__(self, model):
        super().__init__()
        self.convs = model.convs
        self.gru = model.gru
    
    def forward(self, speech):
        xs = speech.unsqueeze(1)
        hs = self.convs(xs).transpose(1, 2)
        new_shape = (hs.size(0), hs.size(1), -1)
        hs = hs.contiguous().view(new_shape)
        _, ref_embs = self.gru(hs)
        return ref_embs[-1]


class OnnxStyleEncoder(nn.Module):
    def __init__(self, model):
        super().__init__()
        self.ref_enc = OnnxReferenceEncoder(model.ref_enc)
        self.stl = model.stl

    def forward(self, speech):
        ref_embs = self.ref_enc(speech)
        style_embs = self.stl(ref_embs)
        return style_embs


class OnnxFastSpeech2(nn.Module, AbsExportModel):
    def __init__(
        self,
        model,
        max_seq_len: int = 512,
        alpha: float = 1.0,
        **kwargs
    ):
        super().__init__()
        # HPs
        self.odim = model.odim
        self.use_gst = model.use_gst
        self.spks = model.spks
        self.langs = model.langs
        self.spk_embed_dim = model.spk_embed_dim
        self.eos = model.eos
        self.model_name = 'fast_speech2'
        if self.spk_embed_dim is not None:
            self.spk_embed_integration_type = model.spk_embed_integration_type
            if self.spk_embed_integration_type == "add":
                self.projection = model.projection
        
        # models
        self.make_pad_mask = MakePadMask(max_seq_len)
        self.encoder = model.encoder
        self.length_regulator = OnnxLengthRegurator(alpha)
        self.pitch_predictor = model.pitch_predictor
        self.energy_predictor = model.energy_predictor
        self.pitch_embed = model.pitch_embed
        self.energy_embed = model.energy_embed
        self.duration_predictor = model.duration_predictor
        self.decoder = model.decoder
        self.feat_out = model.feat_out
        self.postnet = model.postnet
        if self.use_gst:
            self.gst = OnnxStyleEncoder(model.gst)
        if self.spks:
            self.sid_emb = model.sid_emb
        if self.langs:
            self.lid_emb = model.lid_emb
        self.encoder.embed = Embedding(self.encoder.embed, max_seq_len=max_seq_len)
        self.decoder.embed = Embedding(self.decoder.embed, max_seq_len= max_seq_len)

    def _source_mask(self, ilens):
        x_masks = 1 - self.make_pad_mask(ilens)
        return x_masks.unsqueeze(-2)

    def _integrate_with_spk_embed(
        self, hs: torch.Tensor, spembs: torch.Tensor
    ) -> torch.Tensor:
        if self.spk_embed_integration_type == "add":
            # apply projection and then add to hidden states
            spembs = self.projection(F.normalize(spembs))
            hs = hs + spembs.unsqueeze(1)
        elif self.spk_embed_integration_type == "concat":
            # concat hidden states with spk embeds and then apply projection
            spembs = F.normalize(spembs).unsqueeze(1).expand(-1, hs.size(1), -1)
            hs = self.projection(torch.cat([hs, spembs], dim=-1))
        else:
            raise NotImplementedError("support only add or concat.")

        return hs

    def forward(
        self,
        text: torch.Tensor,
        text_length: torch.Tensor,
        feats: Optional[torch.Tensor] = None,
        sids: Optional[torch.Tensor] = None,
        spembs: torch.Tensor = None,
        lids: Optional[torch.Tensor] = None,
    ):
        x, y = text, feats
        spemb = spembs

        # add eos at the last of sequence
        x = F.pad(x, [0, 1], "constant", self.eos)
        text_length += 1

        # setup batch axis
        xs, ys = x.unsqueeze(0), None
        if y is not None:
            ys = y.unsqueeze(0)
        if spemb is not None:
            spembs = spemb.unsqueeze(0)

        _, outs, d_outs, p_outs, e_outs = self._forward(
            xs,
            text_length,
            ys,
            spembs=spembs,
            sids=sids,
            lids=lids,
        )  # (1, T_feats, odim)

        return dict(
            feat_gen=outs[0],
            duration=d_outs[0],
            pitch=p_outs[0],
            energy=e_outs[0],
        )

    def _forward(
        self,
        xs: torch.Tensor,
        ilens: torch.Tensor,
        ys: Optional[torch.Tensor] = None,
        spembs: Optional[torch.Tensor] = None,
        sids: Optional[torch.Tensor] = None,
        lids: Optional[torch.Tensor] = None,
    ):
        x_masks = self._source_mask(ilens)
        hs, _ = self.encoder(xs, x_masks)  # (B, T_text, adim)

        # integrate with GST
        if self.use_gst:
            style_embs = self.gst(ys)
            hs = hs + style_embs.unsqueeze(1)

        # integrate with SID and LID embeddings
        if self.spks is not None:
            sid_embs = self.sid_emb(sids.view(-1))
            hs = hs + sid_embs.unsqueeze(1)
        if self.langs is not None:
            lid_embs = self.lid_emb(lids.view(-1))
            hs = hs + lid_embs.unsqueeze(1)

        # integrate speaker embedding
        if self.spk_embed_dim is not None:
            hs = self._integrate_with_spk_embed(hs, spembs)

        # forward duration predictor and variance predictors
        d_masks = self.make_pad_mask(ilens)

        p_outs = self.pitch_predictor(hs, d_masks.unsqueeze(-1))
        e_outs = self.energy_predictor(hs, d_masks.unsqueeze(-1))

        d_outs = self.duration_predictor.inference(hs, d_masks)  # (B, T_text)
        # use prediction in inference
        p_embs = self.pitch_embed(p_outs.transpose(1, 2)).transpose(1, 2)
        e_embs = self.energy_embed(e_outs.transpose(1, 2)).transpose(1, 2)
        hs = hs + e_embs + p_embs
        hs = self.length_regulator(hs, d_outs)  # (B, T_feats, adim)

        # forward decoder
        h_masks = None
        zs, _ = self.decoder(hs, h_masks)  # (B, T_feats, adim)
        before_outs = self.feat_out(zs).view(
            zs.size(0), -1, self.odim
        )  # (B, T_feats, odim)

        # postnet -> (B, T_feats//r * r, odim)
        if self.postnet is None:
            after_outs = before_outs
        else:
            after_outs = before_outs + self.postnet(
                before_outs.transpose(1, 2)
            ).transpose(1, 2)

        return before_outs, after_outs, d_outs, p_outs, e_outs

    def get_dummy_inputs(self):
        text = torch.LongTensor([0, 1])
        text_length = torch.LongTensor([text.shape[0]])
        feats = torch.randn(10, self.odim) \
            if self.use_gst else None

        sids = torch.LongTensor([0]) \
            if self.spks is not None else None

        spembs = torch.randn(self.spk_embed_dim) \
            if self.spk_embed_dim is not None else None

        lids = torch.LongTensor([0]) \
            if self.langs is not None else None

        return (text, text_length, feats, sids, spembs, lids)

    def get_input_names(self):
        return ['text', 'text_length', 'feats', 'sids', 'spembs', 'lids']

    def get_output_names(self):
        return ['feat_gen', 'out_duration', 'out_pitch', 'out_energy']

    def get_dynamic_axes(self):
        ret = {
            'text': {0: 'text_length'},
            'feats': {0: 'feats_length'},
        }
        return ret

    def get_model_config(self, path):
        return {
            'model_type': 'FastSpeech2',
            'model_path': str(path / f'{self.model_name}.onnx'),
            'eos': self.eos,
        }



from typing import Union
from pathlib import Path
from typeguard import check_argument_types

import os
import glob
from datetime import datetime
import shutil
import logging

import numpy as np
import torch
from onnxruntime.quantization import quantize_dynamic

from espnet2.bin.asr_inference import Speech2Text
from espnet2.text.sentencepiece_tokenizer import SentencepiecesTokenizer
from espnet_model_zoo.downloader import ModelDownloader
from .models import (
    get_encoder,
    get_decoder,
    get_lm,
    RNNDecoder,
    PreDecoder,
    CTC,
    JointNetwork,
)
from .get_config import (
    get_ngram_config,
    get_beam_config,
    get_token_config,
    get_tokenizer_config,
    get_weights_transducer,
    get_trans_beam_config,
)
from espnet_onnx.utils.config import (
    save_config,
    update_model_path
)


class ASRModelExport:
    def __init__(self, cache_dir: Union[Path, str] = None):
        assert check_argument_types()
        if cache_dir is None:
            cache_dir = Path.home() / ".cache" / "espnet_onnx"

        self.cache_dir = Path(cache_dir)
        self.export_config = {}

    def export(
        self,
        model: Speech2Text,
        tag_name: str = None,
        quantize: bool = False,
        verbose: bool = False,
    ):
        assert check_argument_types()
        if tag_name is None:
            tag_name = datetime.now().strftime("%Y%m%d_%H%M%S")
        
        base_dir = self.cache_dir / tag_name.replace(' ', '-')
        export_dir = base_dir / 'full'
        export_dir.mkdir(parents=True, exist_ok=True)

        # copy model files
        self._copy_files(model, base_dir, verbose)

        model_config = self._create_config(model, export_dir)

        # export encoder
        enc_model = get_encoder(model.asr_model.encoder, self.export_config)
        enc_out_size = enc_model.get_output_size()
        self._export_encoder(enc_model, export_dir, verbose)
        model_config.update(encoder=enc_model.get_model_config(
            model.asr_model, export_dir))

        # export decoder
        dec_model = get_decoder(model.asr_model.decoder, self.export_config)
        self._export_decoder(dec_model, enc_out_size, export_dir, verbose)
        model_config.update(decoder=dec_model.get_model_config(export_dir))
        
        # export joint_network if transducer decoder is used.
        if model.asr_model.use_transducer_decoder:
            joint_network = JointNetwork(
                model.asr_model.joint_network,
                model_config['beam_search']['search_type'],
            )
            self._export_joint_network(joint_network, export_dir, verbose)
            model_config.update(joint_network=joint_network.get_model_config(export_dir))

        # export ctc
        ctc_model = CTC(model.asr_model.ctc.ctc_lo)
        self._export_ctc(ctc_model, enc_out_size, export_dir, verbose)
        model_config.update(ctc=ctc_model.get_model_config(export_dir))

        # export lm
        lm_model = None
        if not model.asr_model.use_transducer_decoder:
            if 'lm' in model.beam_search.scorers.keys():
                lm_model = get_lm(model.beam_search.scorers['lm'], self.export_config)
        else:
            if model.beam_search_transducer.use_lm:
                lm_model = get_lm(model.beam_search_transducer.lm, self.export_config)
        
        if lm_model is not None:
            self._export_lm(lm_model, export_dir, verbose)
            model_config.update(lm=lm_model.get_model_config(export_dir))
        else:
            model_config.update(lm=dict(use_lm=False))

        if quantize:
            quantize_dir = base_dir / 'quantize'
            quantize_dir.mkdir(exist_ok=True)
            qt_config = self._quantize_model(export_dir, quantize_dir, verbose)
            for m in qt_config.keys():
                if 'predecoder' in m:
                    model_idx = int(m.split('_')[1])
                    model_config['decoder']['predecoder'][model_idx].update(
                        quantized_model_path=qt_config[m])
                else:
                    model_config[m].update(quantized_model_path=qt_config[m])

        config_name = base_dir / 'config.yaml'
        save_config(model_config, config_name)
        update_model_path(tag_name, base_dir)

    def export_from_pretrained(self, tag_name: str, quantize: bool = False):
        assert check_argument_types()
        model = Speech2Text.from_pretrained(tag_name)
        self.export(model, tag_name, quantize)
    
    def export_from_zip(self, path: Union[Path, str], tag_name: str, quantize: bool = False):
        assert check_argument_types()
        model = Speech2Text.from_pretrained(path)
        self.export(model, tag_name, quantize)
    
    def set_export_config(self, **kwargs):
        for k, v in kwargs.items():
            self.export_config[k] = v

    def _create_config(self, model, path):
        ret = {}
        if not model.asr_model.use_transducer_decoder:
            if "ngram" in list(model.beam_search.full_scorers.keys()) \
                    + list(model.beam_search.part_scorers.keys()):
                ret.update(ngram=get_ngram_config(model))
            else:
                ret.update(ngram=dict(use_ngram=False))
            ret.update(weights=model.beam_search.weights)
            ret.update(beam_search=get_beam_config(
                model.beam_search, model.minlenratio, model.maxlenratio))
        else:
            ret.update(weights=get_weights_transducer(
                model.beam_search_transducer))
            ret.update(beam_search=get_trans_beam_config(
                model.beam_search_transducer
            ))
            
        ret.update(transducer=dict(use_transducer_decoder=model.asr_model.use_transducer_decoder))
        ret.update(token=get_token_config(model.asr_model))
        ret.update(tokenizer=get_tokenizer_config(model.tokenizer, path))
        return ret
    
    def _export_model(self, model, verbose, path, enc_size=None):
        if enc_size:
            dummy_input = model.get_dummy_inputs(enc_size)
        else:
            dummy_input = model.get_dummy_inputs()
            
        torch.onnx.export(
            model,
            dummy_input,
            os.path.join(path, f'{model.model_name}.onnx'),
            verbose=verbose,
            opset_version=15,
            input_names=model.get_input_names(),
            output_names=model.get_output_names(),
            dynamic_axes=model.get_dynamic_axes()
        )
        
        # export submodel if required
        if hasattr(model, 'submodel'):
            for i, sm in enumerate(model.submodel):
                if sm.require_onnx():
                    self._export_model(sm, verbose, path, enc_size)

    def _export_encoder(self, model, path, verbose):
        if verbose:
            logging.info(f'Encoder model is saved in {file_name}')
        self._export_model(model, verbose, path)

    def _export_decoder(self, model, enc_size, path, verbose):
        if verbose:
            logging.info(f'Decoder model is saved in {file_name}')
        self._export_model(model, verbose, path, enc_size)

    def _export_ctc(self, model, enc_size, path, verbose):
        if verbose:
            logging.info(f'CTC model is saved in {file_name}')
        self._export_model(model, verbose, path, enc_size)

    def _export_lm(self, model, path, verbose):
        if verbose:
            logging.info(f'LM model is saved in {file_name}')
        self._export_model(model, verbose, path)
    
    def _export_joint_network(self, model, path, verbose):
        if verbose:
            logging.info(f'JointNetwork model is saved in {file_name}')
        self._export_model(model, verbose, path)
        
    def _copy_files(self, model, path, verbose):
        # copy stats file
        if model.asr_model.normalize is not None \
                and hasattr(model.asr_model.normalize, 'stats_file'):
            stats_file = model.asr_model.normalize.stats_file
            shutil.copy(stats_file, path)
            if verbose:
                logging.info(f'`stats_file` was copied into {path}.')

        # copy bpemodel
        if isinstance(model.tokenizer, SentencepiecesTokenizer):
            bpemodel_file = model.tokenizer.model
            shutil.copy(bpemodel_file, path)
            if verbose:
                logging.info(f'bpemodel was copied into {path}.')
            
        # save position encoder parameters.
        if hasattr(model.asr_model.encoder, 'pos_enc'):
            np.save(
                path / 'pe',
                model.asr_model.encoder.pos_enc.pe.numpy()
            )
            if verbose:
                logging.info(f'Matrix for position encoding was copied into {path}.')

    def _quantize_model(self, model_from, model_to, verbose):
        if verbose:
            logging.info(f'Quantized model is saved in {model_to}.')
        ret = {}
        models = glob.glob(os.path.join(model_from, "*.onnx"))
        for m in models:
            basename = os.path.basename(m).split('.')[0]
            export_file = os.path.join(model_to, basename + '_qt.onnx')
            quantize_dynamic(
                m,
                export_file,
                op_types_to_quantize=['Attention', 'MatMul']
            )
            ret[basename] = export_file
            os.remove(os.path.join(model_from, basename + '-opt.onnx'))
        return ret



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2019 Shigeki Karita
#  Apache 2.0  (http://www.apache.org/licenses/LICENSE-2.0)
# Modified by Masao-Someki
"""Positional Encoding Module."""

import math
import torch
import torch.nn as nn

from espnet.nets.pytorch_backend.transformer.subsampling import Conv2dSubsampling
from espnet.nets.pytorch_backend.transformer.subsampling import Conv2dSubsampling2
from espnet.nets.pytorch_backend.transformer.subsampling import Conv2dSubsampling6
from espnet.nets.pytorch_backend.transformer.subsampling import Conv2dSubsampling8
from espnet.nets.pytorch_backend.transformer.subsampling_without_posenc import Conv2dSubsamplingWOPosEnc
from espnet.nets.pytorch_backend.transformer.embedding import (
    PositionalEncoding,
    ScaledPositionalEncoding,
    RelPositionalEncoding,
    LegacyRelPositionalEncoding,
    StreamPositionalEncoding,
)


def get_pos_emb(pos_emb, max_seq_len=512):
    if isinstance(pos_emb, LegacyRelPositionalEncoding):
        return OnnxLegacyRelPositionalEncoding(pos_emb, max_seq_len)
    elif isinstance(pos_emb, ScaledPositionalEncoding):
        return OnnxScaledPositionalEncoding(pos_emb, max_seq_len)
    elif isinstance(pos_emb, RelPositionalEncoding):
        return OnnxRelPositionalEncoding(pos_emb, max_seq_len)
    elif isinstance(pos_emb, PositionalEncoding):
        return OnnxPositionalEncoding(pos_emb, max_seq_len)
    elif isinstance(pos_emb, StreamPositionalEncoding):
        return OnnxStreamPositionalEncoding(pos_emb, max_seq_len)
    elif (isinstance(pos_emb, nn.Sequential) and len(pos_emb) == 0) \
        or (isinstance(pos_emb, Conv2dSubsamplingWOPosEnc)):
        return pos_emb
    else:
        raise ValueError('Embedding model is not supported.')


class Embedding(nn.Module):
    def __init__(self, model, max_seq_len=512):
        super().__init__()
        self.model = model
        if not isinstance(model, nn.Embedding):
            if (
                isinstance(model, Conv2dSubsampling)
                or isinstance(model, Conv2dSubsampling2)
                or isinstance(model, Conv2dSubsampling6)
                or isinstance(model, Conv2dSubsampling8)
            ):
                self.model.out[-1] = get_pos_emb(model.out[-1], max_seq_len)
            else:
                self.model[-1] = get_pos_emb(model[-1], max_seq_len)

    def forward(self, x, mask=None):
        if mask is None:
            return self.model(x)
        else:
            xs = self.model(x, mask)
            if isinstance(self.model, Conv2dSubsampling):
                return xs, mask[:, :, :-2:2][:, :, :-2:2]
            elif isinstance(self.model, Conv2dSubsampling2):
                return xs, mask[:, :, :-2:2][:, :, :-2:1]
            elif isinstance(self.model, Conv2dSubsampling6):
                return xs, mask[:, :, :-2:2][:, :, :-4:3]
            elif isinstance(self.model, Conv2dSubsampling8):
                return xs, mask[:, :, :-2:2][:, :, :-2:2][:, :, :-2:2]


def _pre_hook(
    state_dict,
    prefix,
    local_metadata,
    strict,
    missing_keys,
    unexpected_keys,
    error_msgs,
):
    """Perform pre-hook in load_state_dict for backward compatibility.

    Note:
        We saved self.pe until v.0.5.2 but we have omitted it later.
        Therefore, we remove the item "pe" from `state_dict` for backward compatibility.

    """
    k = prefix + "pe"
    if k in state_dict:
        state_dict.pop(k)


class OnnxPositionalEncoding(torch.nn.Module):
    """Positional encoding.

    Args:
        d_model (int): Embedding dimension.
        dropout_rate (float): Dropout rate.
        max_seq_len (int): Maximum input length.
        reverse (bool): Whether to reverse the input position. Only for
        the class LegacyRelPositionalEncoding. We remove it in the current
        class RelPositionalEncoding.
    """

    def __init__(self, model, max_seq_len=512, reverse=False):
        """Construct an PositionalEncoding object."""
        super(OnnxPositionalEncoding, self).__init__()
        self.d_model = model.d_model
        self.reverse = model.reverse
        self.xscale = math.sqrt(self.d_model)
        self.pe = None
        self.extend_pe(torch.tensor(0.0).expand(1, max_seq_len))
        self._register_load_state_dict_pre_hook(_pre_hook)

    def extend_pe(self, x):
        """Reset the positional encodings."""
        if self.pe is not None and self.pe.size(1) >= x.size(1):
            if self.pe.dtype != x.dtype or self.pe.device != x.device:
                self.pe = self.pe.to(dtype=x.dtype, device=x.device)
            return
        pe = torch.zeros(x.size(1), self.d_model)
        if self.reverse:
            position = torch.arange(
                x.size(1) - 1, -1, -1.0, dtype=torch.float32
            ).unsqueeze(1)
        else:
            position = torch.arange(
                0, x.size(1), dtype=torch.float32).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, self.d_model, 2, dtype=torch.float32)
            * -(math.log(10000.0) / self.d_model)
        )
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        self.pe = pe.unsqueeze(0)

    def forward(self, x: torch.Tensor):
        """Add positional encoding.

        Args:
            x (torch.Tensor): Input tensor (batch, time, `*`).

        Returns:
            torch.Tensor: Encoded tensor (batch, time, `*`).
        """
        x = x * self.xscale + self.pe[:, : x.size(1)]
        return x


class OnnxScaledPositionalEncoding(OnnxPositionalEncoding):
    """Scaled positional encoding module.

    See Sec. 3.2  https://arxiv.org/abs/1809.08895

    Args:
        d_model (int): Embedding dimension.
        dropout_rate (float): Dropout rate.
        max_seq_len (int): Maximum input length.

    """

    def __init__(self, model, max_seq_len=512):
        """Initialize class."""
        super().__init__(model, max_seq_len)
        self.alpha = torch.nn.Parameter(torch.tensor(1.0))

    def reset_parameters(self):
        """Reset parameters."""
        self.alpha.data = torch.tensor(1.0)

    def forward(self, x):
        """Add positional encoding.

        Args:
            x (torch.Tensor): Input tensor (batch, time, `*`).

        Returns:
            torch.Tensor: Encoded tensor (batch, time, `*`).

        """
        x = x + self.alpha * self.pe[:, : x.size(1)]
        return x


class OnnxLegacyRelPositionalEncoding(OnnxPositionalEncoding):
    """Relative positional encoding module (old version).

    Details can be found in https://github.com/espnet/espnet/pull/2816.

    See : Appendix B in https://arxiv.org/abs/1901.02860

    Args:
        d_model (int): Embedding dimension.
        dropout_rate (float): Dropout rate.
        max_seq_len (int): Maximum input length.

    """

    def __init__(self, model, max_seq_len=512):
        """Initialize class."""
        super().__init__(
            model, max_seq_len,
            reverse=True
        )

    def forward(self, x):
        """Compute positional encoding.

        Args:
            x (torch.Tensor): Input tensor (batch, time, `*`).

        Returns:
            torch.Tensor: Encoded tensor (batch, time, `*`).
            torch.Tensor: Positional embedding tensor (1, time, `*`).

        """
        x = x * self.xscale
        pos_emb = self.pe[:, : x.size(1)]
        return x, pos_emb


class OnnxRelPositionalEncoding(torch.nn.Module):
    """Relative positional encoding module (new implementation).

    Details can be found in https://github.com/espnet/espnet/pull/2816.

    See : Appendix B in https://arxiv.org/abs/1901.02860

    Args:
        d_model (int): Embedding dimension.
        dropout_rate (float): Dropout rate.
        max_seq_len (int): Maximum input length.

    """

    def __init__(self, model, max_seq_len=512):
        """Construct an PositionalEncoding object."""
        super(OnnxRelPositionalEncoding, self).__init__()
        self.d_model = model.d_model
        self.xscale = math.sqrt(self.d_model)
        self.pe = None
        self.extend_pe(torch.tensor(0.0).expand(1, max_seq_len))

    def extend_pe(self, x):
        """Reset the positional encodings."""
        if self.pe is not None and self.pe.size(1) >= x.size(1) * 2 - 1:
            # self.pe contains both positive and negative parts
            # the length of self.pe is 2 * input_len - 1
            if self.pe.dtype != x.dtype or self.pe.device != x.device:
                self.pe = self.pe.to(dtype=x.dtype, device=x.device)
            return
        # Suppose `i` means to the position of query vecotr and `j` means the
        # position of key vector. We use position relative positions when keys
        # are to the left (i>j) and negative relative positions otherwise (i<j).
        pe_positive = torch.zeros(x.size(1), self.d_model)
        pe_negative = torch.zeros(x.size(1), self.d_model)
        position = torch.arange(0, x.size(1), dtype=torch.float32).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, self.d_model, 2, dtype=torch.float32)
            * -(math.log(10000.0) / self.d_model)
        )
        pe_positive[:, 0::2] = torch.sin(position * div_term)
        pe_positive[:, 1::2] = torch.cos(position * div_term)
        pe_negative[:, 0::2] = torch.sin(-1 * position * div_term)
        pe_negative[:, 1::2] = torch.cos(-1 * position * div_term)

        # Reserve the order of positive indices and concat both positive and
        # negative indices. This is used to support the shifting trick
        # as in https://arxiv.org/abs/1901.02860
        pe_positive = torch.flip(pe_positive, [0]).unsqueeze(0)
        pe_negative = pe_negative[1:].unsqueeze(0)
        pe = torch.cat([pe_positive, pe_negative], dim=1)
        self.pe = pe.to(device=x.device, dtype=x.dtype)

    def forward(self, x: torch.Tensor):
        """Add positional encoding.

        Args:
            x (torch.Tensor): Input tensor (batch, time, `*`).

        Returns:
            torch.Tensor: Encoded tensor (batch, time, `*`).

        """
        x = x * self.xscale
        pos_emb = self.pe[
            :,
            self.pe.size(1) // 2 - x.size(1) + 1: self.pe.size(1) // 2 + x.size(1),
        ]
        return x, pos_emb


class OnnxStreamPositionalEncoding(torch.nn.Module):
    """Streaming Positional encoding.

    """
    def __init__(self, model, max_seq_len=5000):
        """Construct an PositionalEncoding object."""
        super(StreamPositionalEncoding, self).__init__()
        
        self.d_model = model.d_model
        self.xscale = model.xscale
        self.pe = None
        # Hold as attribute to export as config parameter,
        # in order to raise an error when start_idx + x.size(1)
        # exceeds the max_seq_len
        tmp = torch.tensor(0.0).expand(1, max_seq_len)
        self.extend_pe(tmp.size(1), tmp.device, tmp.dtype)
        self._register_load_state_dict_pre_hook(_pre_hook)

    def extend_pe(self, length, device, dtype):
        """Reset the positional encodings."""
        pe = torch.zeros(length, self.d_model)
        position = torch.arange(0, length, dtype=torch.float32).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, self.d_model, 2, dtype=torch.float32)
            * -(math.log(10000.0) / self.d_model)
        )
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        self.pe = pe.unsqueeze(0)

    def forward(self, x: torch.Tensor, start_idx: int = 0):
        """Add positional encoding.

        Args:
            x (torch.Tensor): Input tensor (batch, time, `*`).

        Returns:
            torch.Tensor: Encoded tensor (batch, time, `*`).

        """
        return x * self.xscale + self.pe[:, start_idx : start_idx + x.size(1)]



from typing import List
from espnet_onnx.utils.config import Config
from espnet_onnx.asr.model.encoders.encoder import Encoder
from espnet_onnx.asr.model.encoders.streaming import StreamingEncoder


def get_encoder(config: Config, providers: List[str], use_quantized: bool = False):
    if config.enc_type == 'ContextualXformerEncoder':
        return StreamingEncoder(config, providers, use_quantized)
    else:
        return Encoder(config, providers, use_quantized)



from typing import (
    Union,
    Dict,
    List
)
from pathlib import Path
from typeguard import check_argument_types

import numpy as np
import onnxruntime

from .abs_tts_model import AbsTTSModel
from espnet_onnx.tts.model.preprocess.common_processor import CommonPreprocessor
from espnet_onnx.tts.model.duration_calculator import DurationCalculator
from espnet_onnx.tts.model.tts_model import get_tts_model


class Text2Speech(AbsTTSModel):
    """Wrapper class for espnet2.asr.bin.tts_inference.Text2Speech

    """

    def __init__(self,
                 tag_name: str = None,
                 model_dir: Union[Path, str] = None,
                 providers: List[str] = ['CPUExecutionProvider'],
                 use_quantized: bool = False,
                 ):
        assert check_argument_types()
        self._check_argument(tag_name, model_dir)
        self._load_config()

        # check onnxruntime version and providers
        self._check_ort_version(providers)

        # check if there is quantized model if use_quantized=True
        if self.config.tts_model.model_type == 'Tacotron2':
            if use_quantized and 'quantized_model_path' not in self.config.tts_model.encoder.keys():
                raise RuntimeError(
                'Configuration for quantized model is not defined.')
        elif use_quantized and 'quantized_model_path' not in self.config.tts_model.keys():
            # check if quantized model config is defined.
            raise RuntimeError(
                'Configuration for quantized model is not defined.')

        self._build_model(providers, use_quantized)

    def __call__(
        self,
        text: str,
        feats: np.ndarray = None,
        sids: np.ndarray = None,
        spembs: np.ndarray = None,
        lids:  np.ndarray = None,
    ) -> Dict[str, np.ndarray]:
        """Inference

        Args:
            data: Input speech data

        Returns:
            Dict[str, np.ndarray]
        """
        assert check_argument_types()

        # check argument
        options = dict()
        if self.tts_model.use_sids:
            if sids is None or spembs is None:
                raise RuntimeError("'sids' or 'spembs' is missing.")
            else:
                options.update(sids=sids, spembs=spembs)
        if self.tts_model.use_lids:
            if lids is None:
                raise RuntimeError("Missing required argument: 'lids'")
            else:
                options.update(lids=lids)
        if self.tts_model.use_feats:
            if feats is None:
                raise RuntimeError("Missing required argument: 'feats'")
            else:
                options.update(feats=feats)

        # preprocess text
        text = self.preprocess(text)
        output_dict = self.tts_model(text, **options)
        
        # postprocess
        self.postprocess(output_dict['feat_gen'])

        if output_dict.get("att_w") is not None:
            duration, focus_rate = self.duration_calculator(
                output_dict["att_w"])
            output_dict.update(duration=duration, focus_rate=focus_rate)

        # vocoder is currently not supported.
        if self.vocoder is not None:
            input_feat = output_dict["feat_gen"]
            wav = self.vocoder(input_feat)
            output_dict.update(wav=wav)

        return output_dict

    def postprocess(self, feat):
        if self.normalize is not None:
            feat_length = np.array([feat.shape[0]], dtype=np.int64)
            feat, feat_length = self.normalize.inverse(feat[None, :], feat_length)
            return feat[0]
        else:
            return feat




from typing import (
    Optional,
    Tuple
)

import math
import torch
import torch.nn as nn

from espnet_onnx.utils.abs_model import AbsExportModel


class OnnxPWGVocoder(nn.Module, AbsExportModel):
    def __init__(
        self,
        model,
        use_z=False,
        **kwargs
    ):
        super().__init__()
        self.model = model
        self.use_z = use_z
        self.aux_channels = model.aux_channels
        self.model_name = 'PWGVocoder'

    def forward(
        self, c: torch.Tensor, z: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """Perform inference.
        Args:
            c (torch.Tensor): Input tensor (T, in_channels).
            g (Optional[Tensor]): Global conditioning tensor (global_channels, 1).
        Returns:
            Tensor: Output tensor (T ** upsample_factor, out_channels).
        """
        if z is not None:
            z = z.transpose(1, 0).unsqueeze(0)
        c = c.transpose(1, 0).unsqueeze(0)
        return self.model.forward(c, z).squeeze(0).transpose(1, 0)

    def get_dummy_inputs(self):
        c = torch.randn(100, self.aux_channels)
        if self.use_z:
            z = torch.randn(100, self.aux_channels)
            return (c, z)
        else:
            return (c,)

    def get_input_names(self):
        if self.use_z:
            return ['c', 'z']
        else:
            return ['c']

    def get_output_names(self):
        return ['wav']

    def get_dynamic_axes(self):
        ret = {
            'c': {0: 'c_length'},
        }
        if self.use_z:
            ret.update({
                'z': {0: 'z_length'}
            })
        return ret

    def get_model_config(self, path):
        return {
            'vocoder_type': 'OnnxPWGVocoder',
            'model_path': str(path / f'{self.model_name}.onnx')
        }



from typing import Optional
from typing import Tuple
from typeguard import check_argument_types

import numpy as np
import librosa

from espnet_onnx.utils.function import (
    make_pad_mask,
    mask_fill
)
from espnet_onnx.utils.config import Config


class Stft:
    """STFT module.
    """

    def __init__(
        self,
        config: Config
    ):
        assert check_argument_types()
        self.config = config

    def __call__(
        self, input: np.ndarray, ilens: np.ndarray = None
    ) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        """STFT forward function.
        Args:
            input: (Batch, Nsamples)
            ilens: (Batch)
        Returns:
            output: (Batch, Frames, Freq, 2)
        """
        assert check_argument_types()
        stft_kwargs = dict(
            n_fft=self.config.n_fft,
            win_length=self.config.win_length,
            hop_length=self.config.hop_length,
            center=self.config.center,
            window=self.config.window,
        )
        output = []
        # iterate over istances in a batch
        for i, instance in enumerate(input):
            stft = librosa.stft(input[i], **stft_kwargs)
            output.append(np.array(np.stack([stft.real, stft.imag], -1)))
        output = np.vstack(output).reshape(len(output), *output[0].shape)

        if not self.config.onesided:
            len_conj = self.n_fft - output.shape[1]
            conj = output[:, 1: 1 + len_conj].flip(1)
            conj[:, :, :, -1].data *= -1
            output = np.concatenate([output, conj], 1)
        if self.config.normalized:
            output = output * (stft_kwargs["window"].shape[0] ** (-0.5))

        # output: (Batch, Freq, Frames, 2=real_imag)
        # -> (Batch, Frames, Freq, 2=real_imag)
        output = output.transpose(0, 2, 1, 3)
        if ilens is not None:
            if self.config.center:
                pad = self.config.n_fft // 2
                ilens = ilens + 2 * pad
            olens = (ilens - self.config.n_fft) // self.config.hop_length + 1
            output = mask_fill(output, make_pad_mask(
                olens, output, dim=1), 0.0)
        else:
            olens = None
        return output, olens



from typing import Union
from pathlib import Path
from typeguard import check_argument_types

import os
import glob
from datetime import datetime
import logging
import shutil

import numpy as np
import torch
from onnxruntime.quantization import quantize_dynamic

from espnet2.bin.tts_inference import Text2Speech
from espnet_onnx.utils.abs_model import AbsExportModel
from espnet_onnx.export.tts.models import (
    get_tts_model,
    get_vocoder
)
from .get_config import (
    get_token_config,
    get_preprocess_config,
    get_vocoder_config,
    get_normalize_config
)
from espnet_onnx.utils.config import (
    save_config,
    update_model_path
)


class TTSModelExport:
    def __init__(self, cache_dir: Union[Path, str] = None):
        assert check_argument_types()
        if cache_dir is None:
            cache_dir = Path.home() / ".cache" / "espnet_onnx"

        self.cache_dir = Path(cache_dir)
        self.export_config = {}

    def export(
        self,
        model: Text2Speech,
        tag_name: str = None,
        quantize: bool = False,
        verbose: bool = False,
    ):
        assert check_argument_types()
        if tag_name is None:
            tag_name = datetime.now().strftime("%Y%m%d_%H%M%S")

        base_dir = self.cache_dir / tag_name.replace(' ', '-')
        export_dir = base_dir / 'full'
        export_dir.mkdir(parents=True, exist_ok=True)

        # copy model files
        self._copy_files(model, base_dir, verbose)
        
        model_config = self._create_config(model, export_dir)

        # export encoder
        tts_model = get_tts_model(model.model.tts, self.export_config)
        self._export_tts(tts_model, export_dir, verbose)
        if isinstance(tts_model, list):
            model_config.update(tts_model=dict(
                encoder=tts_model[0].get_model_config(export_dir),
                decoder=tts_model[1].get_model_config(export_dir),
                model_type=tts_model[2]
            ))
        else:
            model_config.update(tts_model=tts_model.get_model_config(export_dir))

        # export vocoder
        voc_model, require_export = get_vocoder(model, self.export_config)
        if voc_model is not None:
            if require_export:
                self._export_vocoder(voc_model, export_dir, verbose)
                model_config.update(vocoder=voc_model.get_model_config(export_dir))
            else:
                model_config.update(vocoder=get_vocoder_config(voc_model))

        if quantize:
            quantize_dir = base_dir / 'quantize'
            quantize_dir.mkdir(exist_ok=True)
            qt_config = self._quantize_model(export_dir, quantize_dir, verbose)
            for m in qt_config.keys():
                if 'predecoder' in m:
                    model_config['tts_model']['decoder']['predecoder'].update(
                        quantized_model_path=qt_config[m])
                elif 'postdecoder' in m:
                    model_config['tts_model']['decoder']['postdecoder'].update(
                        quantized_model_path=qt_config[m])
                elif 'tts_model_encoder' in m:
                    model_config['tts_model']['encoder'].update(
                        quantized_model_path=qt_config[m])
                elif 'tts_model_decoder' in m:
                    model_config['tts_model']['decoder'].update(
                        quantized_model_path=qt_config[m])
                else:
                    model_config[m].update(quantized_model_path=qt_config[m])

        config_name = base_dir / 'config.yaml'
        save_config(model_config, config_name)
        update_model_path(tag_name, base_dir)

    def export_from_pretrained(self, tag_name: str, quantize: bool = False):
        assert check_argument_types()
        model = Text2Speech.from_pretrained(tag_name)
        self.export(model, tag_name, quantize)
    
    def export_from_zip(self, path: Union[Path, str], tag_name: str, quantize: bool = False):
        assert check_argument_types()
        model = Text2Speech.from_pretrained(path)
        self.export(model, tag_name, quantize)
    
    def set_export_config(self, **kwargs):
        for k, v in kwargs.items():
            self.export_config[k] = v
    
    def _create_config(self, model, path):
        ret = {}
        ret.update(get_preprocess_config(model.preprocess_fn, path))
        ret.update(normalize=get_normalize_config(model.model.normalize, path))
        ret.update(token=get_token_config(
            model.preprocess_fn.token_id_converter))
        return ret

    def _copy_files(self, model, path, verbose):
        # copy stats file
        if model.model.normalize is not None \
                and hasattr(model.model.normalize, 'stats_file'):
            stats_file = model.model.normalize.stats_file
            shutil.copy(stats_file, path)
            if verbose:
                logging.info(f'`stats_file` was copied into {path}.')

    def _export_model(self, model, verbose, path, enc_size=None):
        if isinstance(model, list):
            for m in model:
                if isinstance(m, AbsExportModel):
                    self._export_model(m, verbose, path, enc_size)
            return
        
        if (
            hasattr(model, 'onnx_export')
            and not model.onnx_export
        ):
            return
        
        if enc_size:
            dummy_input = model.get_dummy_inputs(enc_size)
        else:
            dummy_input = model.get_dummy_inputs()

        torch.onnx.export(
            model,
            dummy_input,
            os.path.join(path, f'{model.model_name}.onnx'),
            verbose=verbose,
            opset_version=15,
            input_names=model.get_input_names(),
            output_names=model.get_output_names(),
            dynamic_axes=model.get_dynamic_axes(),
        )
        
        # export submodel if required
        if hasattr(model, 'submodel'):
            for i, sm in enumerate(model.submodel):
                self._export_model(sm, verbose, path, enc_size)

    def _export_tts(self, model, path, verbose):
        if verbose:
            logging.info(f'TTS model is saved in {file_name}')
        self._export_model(model, verbose, path)

    def _export_vocoder(self, model, path, verbose):
        if verbose:
            logging.info(f'Vocoder model is saved in {file_name}')
        self._export_model(model, verbose, path)

    def _quantize_model(self, model_from, model_to, verbose):
        if verbose:
            logging.info(f'Quantized model is saved in {model_to}.')
        ret = {}
        models = glob.glob(os.path.join(model_from, "*.onnx"))
        for m in models:
            basename = os.path.basename(m).split('.')[0]
            export_file = os.path.join(model_to, basename + '_qt.onnx')
            quantize_dynamic(
                m,
                export_file,
                op_types_to_quantize=['Attention', 'MatMul']
            )
            ret[basename] = export_file
            os.remove(os.path.join(model_from, basename + '-opt.onnx'))
        return ret



from typing import List

import onnxruntime


class JointNetwork:
    def __init__(
        self,
        config,
        providers: List[str],
        use_quantized=False
    ):
        if use_quantized:
            self.joint_session = onnxruntime.InferenceSession(
                config.quantized_model_path,
                providers=providers
            )
        else:
            self.joint_session = onnxruntime.InferenceSession(
                config.model_path,
                providers=providers
            )

    def __call__(self, enc_out, dec_out):
        input_dict = {
            'enc_out': enc_out,
            'dec_out': dec_out
        }
        return self.joint_session.run(['joint_out'], input_dict)[0]



from typing import List
import numpy as np

from espnet_onnx.utils.config import Config

from .batch_beam_search import BatchBeamSearch
from .hyps import Hypothesis


class BatchBeamSearchOnlineSim(BatchBeamSearch):
    """Online beam search implementation.

    This implementation is a modification version
    from espnet.nets.batch_beam_search_online_sim.py
    """

    def set_streaming_config(self, block_size, hop_size, look_ahead):
        """Set config file for streaming decoding.

        Args:
            config (Config): The config for asr training

        """
        self.block_size = block_size
        self.hop_size = hop_size
        self.look_ahead = look_ahead
        self.process_idx = 0
        self.max_frame_len = 10000

    def set_block_size(self, block_size: int):
        """Set block size for streaming decoding.

        Args:
            block_size (int): The block size of encoder
        """
        self.block_size = block_size

    def set_hop_size(self, hop_size: int):
        """Set hop size for streaming decoding.

        Args:
            hop_size (int): The hop size of encoder
        """
        self.hop_size = hop_size

    def set_look_ahead(self, look_ahead: int):
        """Set look ahead size for streaming decoding.

        Args:
            look_ahead (int): The look ahead size of encoder
        """
        self.look_ahead = look_ahead

    def start(self):
        self.conservative = True  # always true
        assert self.block_size and self.hop_size and self.look_ahead, "block_size, hop_size, and look_ahead must be set."
        self.cur_end_frame = int(self.block_size - self.look_ahead)

        # main loop of prefix search
        self.running_hyps = []
        self.ended_hyps = []
        self.prev_repeat = False
        self._init_hyp = True

    def end(self):
        # main loop of prefix search
        self.running_hyps = []
        self.ended_hyps = []
        self.prev_repeat = False
        self._init_hyp = False

    def __call__(self, h: np.ndarray) -> List[Hypothesis]:
        # extend states for ctc
        if self._init_hyp:
            self.minlen = int(self.minlenratio * h.shape[0])
            self.running_hyps = self.init_hyp(h)
            self.prev_hyps = []
            self._init_hyp = False

        maxlen = h.shape[0] if self.maxlenratio == 0 else max(
            1, int(maxlenratio * x.shape[0]))
        move_to_next_block = False
        best = None
        # extend states for ctc
        self.extend(h, self.running_hyps)

        while self.process_idx < h.shape[0]:
            best = self.search(self.running_hyps, h)

            n_batch = best.yseq.shape[0]
            local_ended_hyps = []
            is_local_eos = (
                best.yseq[np.arange(n_batch), best.length - 1] == self.eos
            )
            for i in range(is_local_eos.shape[0]):
                if is_local_eos[i]:
                    hyp = self._select(best, i)
                    local_ended_hyps.append(hyp)
                elif (
                    not self.prev_repeat
                    and best.yseq[i, -1] in best.yseq[i, :-1]
                    and self.cur_end_frame < h.shape[0]
                ):
                    move_to_next_block = True
                    self.prev_repeat = True

            if len(local_ended_hyps) > 0 and self.cur_end_frame < self.max_frame_len:
                move_to_next_block = True

            if move_to_next_block:
                if (
                    self.hop_size
                    and self.cur_end_frame + int(self.hop_size) + int(self.look_ahead)
                    < self.max_frame_len
                ):
                    self.cur_end_frame += int(self.hop_size)
                else:
                    self.cur_end_frame = h.shape[0]

                if self.process_idx > 1 and len(self.prev_hyps) > 0 and self.conservative:
                    self.running_hyps = self.prev_hyps
                    self.process_idx -= 1
                    self.prev_hyps = []
                break

            self.prev_repeat = False
            self.prev_hyps = self.running_hyps
            self.running_hyps = self.post_process(
                self.process_idx, maxlen, best, self.ended_hyps
            )

            if self.cur_end_frame >= h.shape[0]:
                for hyp in local_ended_hyps:
                    self.ended_hyps.append(hyp)

            # increment number
            self.process_idx += 1

        if best is not None:
            return sorted(self.unbatchfy(best), key=lambda x: x.score, reverse=True)
        else:
            return []

    def extend(self, x: np.ndarray, hyps: Hypothesis) -> List[Hypothesis]:
        for k, d in self.scorers.items():
            if hasattr(d, "extend_prob"):
                d.extend_prob(x)
            if hasattr(d, "extend_state"):
                hyps.states[k] = d.extend_state(hyps.states[k])



import os

import torch
import torch.nn as nn

from espnet2.asr.transducer.transducer_decoder import TransducerDecoder

from espnet_onnx.utils.function import subsequent_mask
from ..language_models.embed import Embedding
from espnet_onnx.utils.abs_model import AbsExportModel


class TransducerDecoder(nn.Module, AbsExportModel):
    def __init__(self, model, max_seq_len=512, **kwargs):
        super().__init__()
        self.embed = Embedding(model.embed, max_seq_len)
        self.decoder = model.decoder
        self.dlayers = model.dlayers
        self.dunits = model.dunits
        self.dtype = model.dtype
        self.model_name = 'transducer_encoder'

    def forward(self, labels, h_cache, c_cache):
        # embed and rnn-forward
        sequence = self.embed(labels)
        h_next = torch.zeros(self.dlayers, h_cache.size(1), self.dunits)
        c_next = torch.zeros(self.dlayers, c_cache.size(1), self.dunits)
        if self.dtype == "lstm":
            for i in range(self.dlayers):
                sequence, (h_next[i:i+1], c_next[i:i+1]) = self.decoder[i](
                    sequence,
                    hx=(h_cache[i:i+1], c_cache[i:i+1]) 
                )
        else:
            for i in range(self.dlayers):
                sequence, h_next[i:i+1] = self.decoder[i](
                    sequence, hx=h_cache[i:i+1]
                )
        return sequence, h_next, c_next

    def get_dummy_inputs(self, enc_size):
        labels = torch.LongTensor([0, 1]).unsqueeze(0)
        h_cache = torch.randn(self.dlayers, 1, self.dunits)
        c_cache = torch.randn(self.dlayers, 1, self.dunits)
        return labels, h_cache, c_cache

    def get_input_names(self):
        return ['labels', 'h_cache', 'c_cache'] 

    def get_output_names(self):
        return ['sequence', 'out_h_cache', 'out_c_cache_']
            
    def get_dynamic_axes(self):
        ret = {
            'labels': {
                0: 'labels_batch',
                1: 'labels_length'
            }
        }
        ret.update({
            f'h_cache': {
                1: f'h_cache_length'
            },
            f'c_cache': {
                1: f'c_cache_length'
            },
            f'out_h_cache': {
                1: f'out_h_cache_length'
            },
            f'out_c_cache': {
                1: f'out_c_cache_length'
            }
        })
        return ret

    def get_model_config(self, path):
        return {
            "dec_type": "TransducerDecoder",
            "model_path": os.path.join(path, f'{self.model_name}.onnx'),
            "n_layers": self.dlayers,
            "odim": self.dunits,
            "dtype": self.dtype
        }



import os

import numpy as np
import torch
import torch.nn as nn

from espnet_onnx.utils.abs_model import AbsExportModel


class CTC(nn.Module, AbsExportModel):
    def __init__(self, model):
        super().__init__()
        self.model = model
        self.model_name = 'ctc'

    def forward(self, x):
        return torch.log_softmax(self.model(x), dim=2)

    def get_dummy_inputs(self, enc_size):
        enc_out = torch.randn(1, 100, enc_size)
        return (enc_out,)

    def get_input_names(self):
        return ['x']

    def get_output_names(self):
        return ['ctc_out']

    def get_dynamic_axes(self):
        return {
            "x": {1: "ctc_in_length"},
            "ctc_out": {1: "ctc_out_length"}
        }

    def get_model_config(self, path):
        return {
            "model_path": os.path.join(path, f'{self.model_name}.onnx')
        }



import os

from espnet2.text.char_tokenizer import CharTokenizer
from espnet2.text.phoneme_tokenizer import PhonemeTokenizer
from espnet2.text.sentencepiece_tokenizer import SentencepiecesTokenizer
from espnet2.text.word_tokenizer import WordTokenizer


def get_token_config(model):
    return {
        "list": model.token_list
    }


def get_tokenizer_config(model, path):
    if model is None:
        return {}

    elif isinstance(model, SentencepiecesTokenizer):
        model_name = os.path.basename(model.model)
        return {
            "token_type": "bpe",
            "bpemodel": str(path.parent / model_name)
        }

    elif isinstance(model, WordTokenizer):
        return {
            "token_type": "word"
        }

    elif isinstance(model, CharTokenizer):
        return {
            "token_type": "char"
        }

    elif isinstance(model, PhonemeTokenizer):
        return {
            "token_type": "phn",
            "g2p_type": model.g2p_type
        }


def get_preprocess_config(model, path):
    ret = {}
    if model.text_cleaner is not None:
        ret.update({'text_cleaner': {
                'cleaner_types': model.text_cleaner.cleaner_types[0]
        }})
    else:
        ret.update({'text_cleaner': None})
        
    ret.update({'tokenizer': get_tokenizer_config(model.tokenizer, path)})
    return ret


def get_vocoder_config(model):
    ret = {
        'vocoder_type': 'Spectrogram2Waveform',
    }
    if hasattr(model, 'params'):
        ret.update(model.params)
    return ret


def get_normalize_config(model, path):
    ret = {
        'use_normalize': model is not None
    }
    if ret['use_normalize']:
        ret.update({
            "type": "gmvn",
            "norm_means": model.norm_means,
            "norm_vars": model.norm_vars,
            "eps": model.eps,
            "stats_file": str(path.parent / 'feats_stats.npz')
        })
    return ret


import os

import numpy as np
import torch
import torch.nn as nn

from espnet_onnx.utils.abs_model import AbsExportModel


class JointNetwork(nn.Module, AbsExportModel):
    def __init__(self, model, search_type):
        super().__init__()
        self.model = model
        self.search_type = search_type
        self.model_name = 'joint_network'

    def forward(self, enc_out, dec_out):
        return self.model(enc_out, dec_out)

    def get_dummy_inputs(self):
        if self.search_type in ('default', 'greedy'):
            enc_out = torch.randn(self.model.lin_enc.in_features)
            dec_out = torch.randn(self.model.lin_dec.in_features)
        else:
            enc_out = torch.randn(1, self.model.lin_enc.in_features)
            dec_out = torch.randn(1, self.model.lin_dec.in_features)
        return (enc_out, dec_out)

    def get_input_names(self):
        return ['enc_out', 'dec_out']

    def get_output_names(self):
        return ['joint_out']

    def get_dynamic_axes(self):
        return {
            'enc_out': {
                0: 'enc_out_length'
            },
            'dec_out': {
                0: 'dec_out_length'
            }
        }

    def get_model_config(self, path):
        return {
            "model_path": os.path.join(path, f'{self.model_name}.onnx')
        }



import os

import torch
import torch.nn as nn

from espnet.nets.pytorch_backend.transformer.subsampling import Conv2dSubsampling

from espnet_onnx.utils.function import subsequent_mask
from espnet_onnx.utils.abs_model import AbsExportModel
from .embed import Embedding


class TransformerLM(nn.Module, AbsExportModel):
    def __init__(self, model, max_seq_len=512, **kwargs):
        super().__init__()
        self.embed = Embedding(model.embed, max_seq_len)
        self.encoder = model.encoder
        self.decoder = model.decoder
        self.model_name = 'transformer_lm'

    def forward(self, y, mask, cache):
        xs = self.embed(y)
        # forward_one_step of Encoder
        if isinstance(self.encoder.embed, Conv2dSubsampling):
            xs, mask = self.encoder.embed(xs, mask)
        else:
            xs = self.encoder.embed(xs)

        new_cache = []
        for c, e in zip(cache, self.encoder.encoders):
            xs, mask = e(xs, mask, cache=c)
            new_cache.append(xs)

        if self.encoder.normalize_before:
            xs = self.encoder.after_norm(xs)

        h = self.decoder(xs[:, -1])
        return h, new_cache

    def get_dummy_inputs(self):
        tgt = torch.LongTensor([0, 1]).unsqueeze(0)
        ys_mask = tgt != 0
        m = torch.from_numpy(subsequent_mask(ys_mask.shape[-1])[None, :])
        mask = ys_mask[None, :] * m
        cache = [
            torch.zeros((1, 1, self.encoder.encoders[0].size))
            for _ in range(len(self.encoder.encoders))
        ]
        return (tgt, mask, cache)

    def get_input_names(self):
        return ['tgt', 'tgt_mask'] \
            + ['cache_%d' % i for i in range(len(self.encoder.encoders))]

    def get_output_names(self):
        return ['y'] \
            + ['out_cache_%d' % i for i in range(len(self.encoder.encoders))]

    def get_dynamic_axes(self):
        ret = {
            'tgt': {
                0: 'tgt_batch',
                1: 'tgt_length'
            },
            'tgt_mask': {
                0: 'tgt_mask_batch',
                1: 'tgt_mask_length',
                2: 'tgt_mask_height'
            }
        }
        ret.update({
            'cache_%d' % d: {
                0: 'cache_%d_batch' % d,
                1: 'cache_%d_length' % d
            }
            for d in range(len(self.encoder.encoders))
        })
        ret.update({
            'out_cache_%d' % d: {
                0: 'out_cache_%d_batch' % d,
                1: 'out_cache_%d_length' % d
            }
            for d in range(len(self.encoder.encoders))
        })
        return ret

    def get_model_config(self, path):
        return {
            "use_lm": True,
            "model_path": os.path.join(path, f'{self.model_name}.onnx'),
            "lm_type": "TransformerLM",
            "odim": self.encoder.encoders[0].size,
            "nlayers": len(self.encoder.encoders)
        }
